from rest_framework.routers import DefaultRouter

from django_project_base.notifications.rest.delivery_report import DeliveryReportViewSet
from django_project_base.notifications.rest.maintenance_notification import UsersMaintenanceNotificationViewset
from django_project_base.notifications.rest.notification import NotificationsLicenseViewSet, NotificationViewset


class NotificationsRouter(DefaultRouter):
    pass


notifications_router: NotificationsRouter = DefaultRouter()

notifications_router.register(
    r"maintenance-notification", UsersMaintenanceNotificationViewset, basename="maintenance-notification"
)
notifications_router.register(r"notification", NotificationViewset, basename="notification")
notifications_router.register(r"notification-license", NotificationsLicenseViewSet, basename="notification-license")
notifications_router.register(r"notification-dlr", DeliveryReportViewSet, basename="notification-dlr", )



from enum import IntEnum
from typing import Dict, TYPE_CHECKING, Union

from django.db.models import Model, QuerySet
from django.http import HttpRequest
from rest_framework.utils.model_meta import get_field_info

if TYPE_CHECKING:
    from dynamicforms.serializers import Serializer


def set_django_security(django_settings, deploy=True, swagger_version=None):
    """
    Sets django web security settings.
    Call this from settings.py

    :param django_settings: globals() of your application
    :param deploy: True if server is deployed, False for development
    :param swagger_version: If you are using swagger or drf_spectacular set swagger version to allow CSP to access
        swagger files.
    :return:
    """

    for app in ["django_permissions_policy", "csp"]:
        if app not in django_settings["INSTALLED_APPS"]:
            django_settings["INSTALLED_APPS"].append(app)

    for middleware in ["django_permissions_policy.PermissionsPolicyMiddleware", "csp.middleware.CSPMiddleware"]:
        if middleware not in django_settings["MIDDLEWARE"]:
            django_settings["MIDDLEWARE"].append(middleware)

    django_settings["SECURE_BROWSER_XSS_FILTER"] = True
    django_settings["SECURE_CONTENT_TYPE_NOSNIFF"] = True

    if deploy:
        django_settings["SECURE_SSL_REDIRECT"] = True
        django_settings["SECURE_PROXY_SSL_HEADER"] = ("HTTP_X_FORWARDED_PROTO", "https")

        django_settings["CSRF_COOKIE_SECURE"] = True
        django_settings["CSRF_COOKIE_NAME"] = "__Host-csrftoken"
        django_settings["SESSION_COOKIE_SECURE"] = True
        django_settings["SESSION_COOKIE_NAME"] = "__Host-sessionid"

        django_settings["SECURE_HSTS_SECONDS"] = 15768000  # 6 Months
        django_settings["SECURE_HSTS_INCLUDE_SUBDOMAINS"] = True
        django_settings["SECURE_HSTS_PRELOAD"] = True

    django_settings["CSRF_COOKIE_HTTPONLY"] = True
    django_settings["SESSION_COOKIE_HTTPONLY"] = True

    django_settings["CSRF_COOKIE_SAMESITE"] = "Strict"
    django_settings["SESSION_COOKIE_SAMESITE"] = "Strict"

    # CSP can be changed per request... with decorator:
    #  https://django-csp.readthedocs.io/en/latest/decorators.html#decorator-chapter
    # Content Security Policy
    django_settings["CSP_DEFAULT_SRC"] = [
        "'none'" if deploy else "'self'",
    ]
    # stackpath.bootstrapcdn.com - because of (old) login
    django_settings["CSP_STYLE_SRC"] = [
        "'self'",
        "'unsafe-inline'",
        "cdnjs.cloudflare.com",
        "stackpath.bootstrapcdn.com",
        "cdn.jsdelivr.net/npm/@mdi/font@latest/",
    ]
    # unsafe-eval - because of Vue.js
    # code.jquery.com, stackpath.bootstrapcdn.com, connect.facebook.net - because of (old) login
    django_settings["CSP_SCRIPT_SRC"] = [
        "'self'",
        "'unsafe-inline'",
        "'unsafe-eval'",
        "code.jquery.com",
        "stackpath.bootstrapcdn.com",
        "connect.facebook.net",
    ]
    # via.placeholder.com - because of default user logo
    django_settings["CSP_IMG_SRC"] = ["'self'", "via.placeholder.com"]

    if swagger_version:
        django_settings["CSP_STYLE_SRC"].append("cdn.jsdelivr.net/npm/swagger-ui-dist@%s/" % swagger_version)
        django_settings["CSP_SCRIPT_SRC"].append("cdn.jsdelivr.net/npm/swagger-ui-dist@%s/" % swagger_version)
        django_settings["CSP_IMG_SRC"].append("cdn.jsdelivr.net/npm/swagger-ui-dist@%s/" % swagger_version)

    django_settings["CSP_FONT_SRC"] = ["'self'", "cdnjs.cloudflare.com", "cdn.jsdelivr.net/npm/@mdi/font@latest/"]
    django_settings["CSP_CONNECT_SRC"] = ["'self'", "unpkg.com/ionicons@5.5.1/"]
    django_settings["CSP_OBJECT_SRC"] = [
        "'none'",
    ]
    django_settings["CSP_BASE_URI"] = [
        "'self'",
    ]
    # TODO temp removed because we're rendering all legacy pages as iframes when everything is moved to vue, reenable
    # django_settings["CSP_FRAME_ANCESTORS"] = [
    #     "'none'",
    # ]
    django_settings["CSP_FORM_ACTION"] = [
        "'self'",
    ]

    django_settings["PERMISSIONS_POLICY"] = {
        "accelerometer": [],
        "ambient-light-sensor": [],
        "autoplay": [],
        # "battery": [],
        "camera": [],
        "display-capture": [],
        "document-domain": [],
        "encrypted-media": [],
        "fullscreen": [],
        "gamepad": [],
        "geolocation": [],
        "gyroscope": [],
        "interest-cohort": [],
        "magnetometer": [],
        "microphone": [],
        "midi": [],
        "payment": [],
        "picture-in-picture": [],
        "publickey-credentials-get": [],
        "screen-wake-lock": [],
        # "speaker-selection": [],
        "sync-xhr": [],
        "usb": [],
        "web-share": [],
        "xr-spatial-tracking": [],
    }


class IntDescribedEnum(IntEnum):
    def __new__(cls, *args, **kwds):
        obj = int.__new__(cls, args[0])
        obj._value_ = args[0]
        return obj

    # ignore the first param since it's already set by new
    def __init__(self, _: str, description: str = None, data: dict = None):
        self._description_ = description
        self._data_ = data

    description = property(lambda self: self._description_)
    data = property(lambda self: self._data_)

    @classmethod
    def get_choices_tuple(cls):
        return tuple((item.value, item.description) for item in cls)

    @classmethod
    def has_value(cls, value):
        return value in cls.__members__.values()


def get_pk_name(obj: Union[Model, QuerySet, Dict, "Serializer"], model: Model = None):
    if isinstance(obj, QuerySet):
        return obj.model._meta.pk.name
    if isinstance(obj, Model):
        return obj._meta.pk.name
    return get_field_info(model if model else obj).pk.name


def get_host_url(request: HttpRequest) -> str:
    host_url = "%s://%s" % ("https" if request.is_secure() else "http", request.META["HTTP_HOST"])
    if not host_url.endswith("/"):
        host_url += "/"
    return host_url



"""
A ModelViewSet base to provide automatic filtering based on currently selected project
"""

from dynamicforms.viewsets import ModelViewSet

from django_project_base.base.filter_to_model import filter_queryset_or_model_to_project


class ProjectFilteringViewSet(ModelViewSet):
    """
    ViewSet that automatically performs filtering on project for the given model

    You may provide a queryset, just like in standard DRF ViewSets
    You may as well provide a model class that will be used instead of the queryset.
    The model instance may also be a callable that will return a model class when called (for deferred resolution)
    """

    model = None
    project_field = None

    def get_queryset(self):
        return filter_queryset_or_model_to_project(
            self.queryset, self.model, self.project_field or "project", self.request.selected_project
        )



import uuid

from django.conf import settings

from django_project_base.licensing.logic import LogAccessService
from django_project_base.notifications.base.channels.channel import Recipient
from django_project_base.notifications.base.enums import ChannelIdentifier
from django_project_base.notifications.base.notification import Notification
from django_project_base.notifications.models import DeliveryReport, DjangoProjectBaseNotification
from django_project_base.notifications.tests.notifications_transaction_test_case import (
    NotificationsTransactionTestCase,
    TestNotificationViaEmail,
)


class IsEmailSentTest(NotificationsTransactionTestCase):
    def test_is_email_sent(self):
        notification: TestNotificationViaEmail = self.create_notification_email_object()
        notification.send()
        notification_db_object: DjangoProjectBaseNotification = DjangoProjectBaseNotification.objects.all().first()
        self.assertEqual(len(notification_db_object.sent_channels.split(",")), 1)
        assert isinstance(notification_db_object.sent_at, int)
        self.assertIsNone(notification_db_object.exceptions)

    def test_resend_email(self):
        lic_start = LogAccessService().report(self.test_user)
        notification: DjangoProjectBaseNotification = self.create_notification_email_object().send()
        lic_first_sent = LogAccessService().report(self.test_user)
        self.assertIsNotNone(notification.sent_at)
        self.assertFalse(bool(notification.failed_channels))
        self.assertEqual(notification.required_channels, notification.sent_channels)
        channel = ChannelIdentifier.channel(
            notification.required_channels.split(",")[0], ensure_dlr_user=False, settings=settings
        )
        dlr_pk = str(uuid.uuid4())
        dlr_count = DeliveryReport.objects.all().count()
        resend_data = channel._make_send(
            notification_obj=notification,
            rec_obj=Recipient(identifier=str(self.test_user.pk), phone_number="+38634000000", email="xyz@xyz.xyz"),
            message_str="message",
            dlr_pk=dlr_pk,
        )
        self.assertFalse(resend_data[1])
        self.assertIsNone(resend_data[0])
        self.assertEqual(dlr_count, DeliveryReport.objects.all().count())
        Notification.resend(notification=notification, user_pk=self.test_user.pk)
        self.assertEqual(dlr_count, DeliveryReport.objects.all().count())
        lic_resend = LogAccessService().report(self.test_user)
        self.assertEqual(lic_first_sent["remaining_credit"], lic_resend["remaining_credit"])
        self.assertTrue(lic_start["remaining_credit"] > lic_resend["remaining_credit"])

    # def test_is_email_queued(self):
    #     mail_queue: Queue = QueueFactory("mail")
    #     mail_queue.clear()
    #     send_mail("test", "test", device=None, to_mail=["aa@ff.si"], mail_content_entity_context={})
    #     self.assertEqual(1, mail_queue.length)



import json
import re

from io import BytesIO

import pandas as pd
import swapper

from django.contrib.auth.hashers import make_password
from django.db import transaction
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _
from dynamicforms import fields, serializers
from dynamicforms.action import TableAction, TablePosition
from dynamicforms.viewsets import SingleRecordViewSet
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.utils import model_meta

from django_project_base.account.rest.profile import ProfileSerializer, ProfileViewSet

from ...base.permissions import is_project_owner, is_superuser, IsProjectOwnerOrMemberReadOnly, project_is_selected
from ..middleware import ProjectNotSelectedError
from .project_profiles_utils import filter_project_members_fields, get_project_members


class ProjectProfilesSerializer(ProfileSerializer):
    template_context = dict(url_reverse="alc-profiles")

    def __init__(self, *args, **kwargs):
        request = kwargs["context"]["request"]
        self.member_fields_read_only = bool(re.findall(r"profile/current", request.path))

        user = request.user
        project = request.selected_project

        self.user_is_admin = is_superuser(user) or (project_is_selected(project) and is_project_owner(user, project))
        self.user_is_me = (
            len(args) == 1
            and isinstance(args[0], swapper.load_model("django_project_base", "Profile"))
            and user.id == args[0].id
        )

        super().__init__(*args, **kwargs)

        if self.user_is_admin:
            self.actions.actions.extend(
                [
                    TableAction(
                        TablePosition.ROW_END,
                        label=_("Reset password"),
                        name="reset-password",
                        icon="refresh-outline",
                    ),
                    TableAction(
                        TablePosition.ROW_END,
                        label=_("Merge"),
                        name="add-to-merge",
                        icon="git-merge-outline",
                    ),
                ]
            )

    def get_fields(self):
        ProjectMember = swapper.load_model("django_project_base", "ProjectMember")
        info = model_meta.get_field_info(ProjectMember)
        for field in ProjectMember().project_members_fields:
            source = f"projects.first.{field.name}"
            extra_kwargs = dict(source=source, required=False, read_only=self.member_fields_read_only)
            field_class, field_kwargs = self.build_field(field.name, info, ProjectMember, source)
            field_kwargs = self.include_extra_kwargs(field_kwargs, extra_kwargs)
            self._declared_fields[field.name] = field_class(**field_kwargs)
        res = super().get_fields()

        if not self.user_is_admin and not self.user_is_me:
            # if user is not super admin make all fields readonly
            for field in res.values():
                if getattr(field, "read_only", None) is not None:
                    setattr(field, "read_only", True)

        if self.user_is_admin:
            password_field = res.get("password", None)
            if password_field:
                setattr(password_field, "display", fields.DisplayMode.FULL)
                setattr(password_field, "display_table", fields.DisplayMode.HIDDEN)

        return res

    def translate_relation_fields(self, field_name: str) -> str:
        ProjectMember = swapper.load_model("django_project_base", "ProjectMember")
        if field_name in ProjectMember().project_members_fields_names:
            return f"projects__{field_name}"
        else:
            return field_name

    def get_visible_fields(self):
        return [(name, field) for (name, field) in self.fields.items() if field.display == fields.DisplayMode.FULL]


class ProjectProfilesViewSet(ProfileViewSet):
    serializer_class = ProjectProfilesSerializer
    # see special case in the permission class if you need to change the permission class
    permission_classes = (IsProjectOwnerOrMemberReadOnly,)

    def get_serializer_class(self):
        return ProjectProfilesSerializer

    def filter_queryset_field(self, queryset, field, value):
        if field == "state" and value:
            return queryset.filter(projects__state=value)
        return super().filter_queryset_field(queryset, field, value)

    def save_club_member_data(self, request: Request, user, **kwargs):
        if user is None:
            return
        project = request.selected_project

        club_member = None
        ProjectMember = swapper.load_model("django_project_base", "ProjectMember")

        if project:
            club_member = ProjectMember.objects.filter(member=user).filter(project=project).first()

        # if club member data cant be retrieved, we can't save anything
        if club_member:
            for name, value in kwargs.items():
                setattr(club_member, name, value)
            club_member.save()

    @transaction.atomic
    def create(self, request: Request, *args, **kwargs) -> Response:
        ProjectMember = swapper.load_model("django_project_base", "ProjectMember")
        Profile = swapper.load_model("django_project_base", "Profile")
        data = {
            name: request.data.pop(name, None)
            for name in ProjectMember().project_members_fields_names
            if name in request.data
        }
        # hash password if it exists
        if "password" in request.data:
            request.data["password"] = make_password(request.data["password"])
        response = super().create(request, *args, **kwargs)
        user = Profile.objects.get(pk=response.data["id"])
        try:
            ProjectMember.objects.create(project=request.selected_project, member=user)
        except ProjectNotSelectedError as e:
            raise PermissionDenied(e.message)

        self.save_club_member_data(request, user, **data)
        return response

    @transaction.atomic
    def update(self, request: Request, *args, **kwargs) -> Response:
        ProjectMember = swapper.load_model("django_project_base", "ProjectMember")
        data = {
            name: request.data.pop(name, None)
            for name in ProjectMember().project_members_fields_names
            if name in request.data
        }
        self.save_club_member_data(request, self.get_object(), **data)
        # hash password if it exists
        if "password" in request.data:
            request.data["password"] = make_password(request.data["password"])
        return super().update(request, *args, **kwargs)


class ProfileExportSerializer(serializers.Serializer):
    template_context = dict(url_reverse="alc-profiles-export")

    form_titles = {
        "table": "",
        "new": _("Export users"),
        "edit": "",
    }

    template = fields.ChoiceField(
        label=_("Template"),
        choices=(
            # (1, _("PDF")),
            (2, _("CSV")),
            (3, _("XLS")),
        ),
    )

    profile_fields = fields.MultipleChoiceField(label=_("Profile Fields"), allow_null=True)

    filter_data = fields.JSONField(display=fields.DisplayMode.HIDDEN)

    print_filter = fields.BooleanField(label=_("Print Filter"), default=False)

    def __init__(self, *args, **kwargs):
        profile_serializer = ProjectProfilesSerializer(None, **kwargs)
        visible_profile_fields = [(name, field.label) for name, field in profile_serializer.get_visible_fields()]
        super().__init__(*args, **kwargs)
        self.fields["profile_fields"].choices = visible_profile_fields


class ProfileExportViewSet(SingleRecordViewSet):
    serializer_class = ProfileExportSerializer

    def get_permissions(self):
        if self.action == "download":
            return [IsAuthenticated()]
        else:
            return super().get_permissions()

    def new_object(self):
        filter_data = self.request.query_params.get("filter_data", None)
        if filter_data:
            filter_data = json.loads(filter_data)
        return dict(template=None, profile_fields=None, print_filter=True, filter_data=filter_data)

    def create(self, request, *args, **kwargs):
        ser = self.get_serializer(data=request.data)  # type: ProfileExportSerializer
        ser.is_valid(raise_exception=True)

        filter_data = ser.data.get("filter_data", None)
        print_filter = ser.data.get("print_filter", False)

        file_name = "members"

        if filter_data:
            active_filters = {name: value for name, value in filter_data.items() if value is not None}
            if print_filter and len(active_filters) in [1, 2]:
                file_name += "-" + ",".join(["-".join([name, str(value)]) for name, value in active_filters.items()])

        data = dict(ser.data)
        data["file_name"] = file_name

        return Response(data)

    @action(methods=["POST"], detail=False, url_path="download", url_name="profile-export-download")
    def download(self, request: Request, **kwargs):
        profile_serializer = ProjectProfilesSerializer(None, context=self.get_serializer_context(), data=request.data)
        visible_profile_fields = {
            name: profile_serializer.translate_relation_fields(name)
            for name, _ in profile_serializer.get_visible_fields()
        }
        column_names = {
            visible_profile_fields[name]: field.label for name, field in profile_serializer.get_visible_fields()
        }

        ser = self.get_serializer(data=request.data)  # type: ProfileExportSerializer
        ser.is_valid(raise_exception=True)

        profile_fields = ser.data.get("profile_fields", None)
        template = ser.data.get("template", None)
        filter_data = ser.data.get("filter_data", None)
        file_name = ser.data.get("file_name", "members")

        profile_items_q = get_project_members(self.request)

        if filter_data:
            for field_name, value in filter_data.items():
                profile_items_q = filter_project_members_fields(profile_items_q, field_name, value)

        selected_fields = (
            [name for name in visible_profile_fields.values()]
            if profile_fields is None
            else [visible_profile_fields[field] for field in profile_fields if field in visible_profile_fields]
        )

        profile_items_q = profile_items_q.values(*selected_fields)

        output = BytesIO()
        df = pd.DataFrame(profile_items_q, columns=list(selected_fields))
        df.rename(columns=column_names, inplace=True)

        if template == 2:
            df.to_csv(output)
            output.seek(0)
            response = HttpResponse(output, content_type="text/csv")
            response["Content-Disposition"] = f'attachment; filename="{file_name}.csv"'
            response["Content-Name"] = f"{file_name}.csv"
            return response
        elif template == 3:
            df.to_excel(output)
            output.seek(0)
            response = HttpResponse(output, content_type="application/vnd.ms-excel")
            response["Content-Disposition"] = f'attachment; filename="{file_name}.xlsx"'
            response["Content-Name"] = f"{file_name}.xlsx"
            return response
        else:
            raise NotFound(f"Export template {template} not found")



SOCIAL_AUTH_SETTINGS = (
    {
        "name": "AUTHENTICATION_BACKENDS",
        "default": (
            "social_core.backends.facebook.FacebookOAuth2",
            "social_core.backends.google.GoogleOAuth2",
            "social_core.backends.twitter.TwitterOAuth",
            "social_core.backends.microsoft.MicrosoftOAuth2",
            "social_core.backends.github.GithubOAuth2",
            "social_core.backends.gitlab.GitLabOAuth2",
            # 'social_core.backends.apple.AppleIdAuth', # not fully tested yet
        ),
        "description": "Social login authentication backends. Add desired authentication backends to "
        "Django’s AUTHENTICATION_BACKENDS setting.",
    },
    {
        "name": "SOCIAL_AUTH_REDIRECT_IS_HTTPS",
        "default": False,
        "description": "On projects behind a reverse proxy that uses HTTPS, the redirect URIs can have the wrong "
        "schema (http:// instead of https://) if the request lacks the appropriate headers, which might "
        "cause errors during the auth process. To force HTTPS in the "
        "final URIs set this setting to True",
    },
    {
        "name": "SOCIAL_AUTH_LOGIN_REDIRECT_URL",
        "default": "/",
        "description": "Used to redirect the user once the auth process ended successfully. "
        "The value of ?next=/foo is used if it was present",
    },
    {
        "name": "SOCIAL_AUTH_PIPELINE",
        "default": (
            "social_core.pipeline.social_auth.social_details",
            "social_core.pipeline.social_auth.social_uid",
            "social_core.pipeline.social_auth.auth_allowed",
            "social_core.pipeline.social_auth.social_user",
            "social_core.pipeline.user.get_username",
            "social_core.pipeline.social_auth.associate_by_email",
            "social_core.pipeline.user.create_user",
            "social_core.pipeline.social_auth.associate_user",
            "social_core.pipeline.social_auth.load_extra_data",
            "social_core.pipeline.user.user_details",
        ),
        "description": "python-social-auth authentication workflow pipeline. The default pipeline is a mechanism "
        "that creates user instances and gathers basic data from providers.",
    },
)



import json

import swapper

from django.conf import settings

from django_project_base.notifications.email_notification import SystemEMailNotificationWithListOfEmails
from django_project_base.notifications.models import DjangoProjectBaseMessage
from django_project_base.profiling.performance_base_command import PerformanceCommand


class Command(PerformanceCommand):
    help = "Lists pending project settings. Example:  python manage.py list_pending_settings"

    def handle(self, *args, **options):
        result = dict()
        for project in swapper.load_model("django_project_base", "Project").objects.all():
            for setting in (
                swapper.load_model("django_project_base", "ProjectSettings")
                .objects.filter(project=project, pending_value__isnull=False)
                .exclude(pending_value="")
            ):
                if project.name not in result:
                    result[project.name] = {}
                result[project.name][setting.name] = {
                    "value": setting.python_value,
                    "pending_value": setting.python_pending_value,
                }

        if to := getattr(settings, "ADMINS", getattr(settings, "MANAGERS", [])) and result:
            SystemEMailNotificationWithListOfEmails(
                message=DjangoProjectBaseMessage(
                    subject="Pending settings report",
                    body=json.dumps(result),
                    footer="",
                    content_type=DjangoProjectBaseMessage.HTML,
                ),
                recipients=to,
            ).send()

        self.stdout.write(self.style.WARNING(result))



import datetime
import json
import time

from typing import List, Optional

import pytz
import swapper

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.db.models import ForeignKey
from django.http import Http404, HttpResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema
from dynamicforms import fields
from dynamicforms.action import Actions, FormButtonAction, FormButtonTypes, TableAction, TablePosition
from dynamicforms.mixins import DisplayMode, F
from dynamicforms.mixins.conditional_visibility import Operators, Statement
from dynamicforms.serializers import ModelSerializer, Serializer
from dynamicforms.template_render.layout import Column, Layout, Row
from dynamicforms.template_render.responsive_table_layout import ResponsiveTableLayout, ResponsiveTableLayouts
from dynamicforms.viewsets import ModelViewSet, SingleRecordViewSet
from rest_framework import status
from rest_framework.authentication import BasicAuthentication, SessionAuthentication, TokenAuthentication
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.fields import empty
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response

from django_project_base.account.middleware import ProjectNotSelectedError
from django_project_base.licensing.logic import LicenseReportSerializer, LogAccessService
from django_project_base.notifications.base.channels.sms_channel import SmsChannel
from django_project_base.notifications.base.enums import ChannelIdentifier
from django_project_base.notifications.base.notification import Notification
from django_project_base.notifications.models import (
    DjangoProjectBaseMessage,
    DjangoProjectBaseNotification,
    SearchItems,
)
from django_project_base.utils import get_host_url, get_pk_name


class MessageBodyField(fields.RTFField):
    def __init__(self, *args, **kw):
        kw["write_only"] = True
        kw["label"] = _("Body")
        kw["display_table"] = DisplayMode.SUPPRESS
        super().__init__(*args, **kw)
        # TODO: if field is write only and not present in model serializer table, field is not rendered
        self.render_params["form_component_name"] = "DCKEditor"


class OrginalRecipientsField(fields.CharField):
    def to_representation(self, value, row_data=None):
        if row_data and row_data.recipients_original_payload_search:
            if len(row_data.recipients_original_payload_search) > 95:
                return f"{row_data.recipients_original_payload_search[:95]} ..."
            return row_data.recipients_original_payload_search
        if value:
            search_str = ",".join(
                list(
                    map(
                        str,
                        MessageToListField().parse(val=json.loads(value), return_instances=True),
                    )
                )
            )
            instance = self.parent.instance if self.parent else None
            if (
                instance
                and isinstance(instance, DjangoProjectBaseNotification)
                and not instance.recipients_original_payload_search
            ):
                instance.recipients_original_payload_search = search_str
                instance.save(update_fields=["recipients_original_payload_search"])
            # TODO: THIS SOLUTION FOR SEARCH IS BAD; BAD; -> MAKE BETTER ONE
            if isinstance(row_data, DjangoProjectBaseNotification) and not row_data.recipients_original_payload_search:
                row_data.recipients_original_payload_search = search_str
                row_data.save(update_fields=["recipients_original_payload_search"])
            if len(search_str) > 95:
                # TODO: INITIAL ROWS IN TABLE RENDER AND NOT HANDLED BY RENDER TO TABLE
                search_str = f"{search_str[:95]} ..."
            return search_str
        return super().to_representation(value, row_data)

    def render_to_table(self, value, row_data):
        val = super().render_to_table(value=value, row_data=row_data)
        if len(val) > 95:
            val = f"{val[:95]} ..."
        return val


class ReadOnlyDateTimeFieldFromTs(fields.DateTimeField):
    def to_representation(self, value, row_data=None):
        if value:
            return datetime.datetime.fromtimestamp(value).astimezone(pytz.utc)
        return value


class NotificationSerializer(ModelSerializer):
    template_context = dict(url_reverse="notification")
    form_titles = {"table": _("Messages"), "edit": _("New message")}

    def __init__(self, *args, is_filter: bool = False, **kwds):
        super().__init__(*args, is_filter=is_filter, **kwds)
        self.fields.fields["level"].display_form = DisplayMode.HIDDEN
        self.fields.fields["type"].display_form = DisplayMode.HIDDEN
        self.fields.fields["project_slug"].display = DisplayMode.HIDDEN
        self.fields.fields["message_to"].child_relation.queryset = SearchItems.objects.get_queryset(
            request=self.request
        )
        self.fields.fields["send_notification_sms_text"].display = DisplayMode.SUPPRESS

    id = fields.UUIDField(display=DisplayMode.HIDDEN)

    subject = fields.SerializerMethodField(display_form=DisplayMode.HIDDEN)
    recipients = fields.CharField(display_form=DisplayMode.HIDDEN, display_table=DisplayMode.HIDDEN)

    recipients_original_payload = OrginalRecipientsField(
        display_form=DisplayMode.HIDDEN, label=_("Recipients"), read_only=True, render_params={"max_width": "20em"}
    )

    delivery = fields.SerializerMethodField(label=_("Delivery"))

    level = fields.CharField(display=DisplayMode.SUPPRESS)
    type = fields.CharField(display=DisplayMode.SUPPRESS)

    message = fields.PrimaryKeyRelatedField(
        display_form=DisplayMode.SUPPRESS,
        display_table=DisplayMode.SUPPRESS,
        read_only=True,
    )

    actions = Actions(
        TableAction(
            TablePosition.HEADER,
            label=_("Add"),
            title=_("Add new notification"),
            name="add-notification",
            icon="add-circle-outline",
        ),
        TableAction(
            TablePosition.HEADER,
            label=_("View license"),
            title=_("View license"),
            name="view-license",
            icon="card-outline",
        ),
        TableAction(TablePosition.ROW_CLICK, _("Edit"), title=_("Edit record"), name="edit", icon="pencil-outline"),
        FormButtonAction(
            btn_type=FormButtonTypes.CANCEL,
            name="cancel",
        ),
        FormButtonAction(
            btn_type=FormButtonTypes.SUBMIT,
            label=_("Send"),
            name="submit",
        ),
        add_form_buttons=False,
    )

    message_to = fields.ManyRelatedField(
        child_relation=fields.PrimaryKeyRelatedField(
            queryset=SearchItems.objects.get_queryset(),
            required=True,
        ),
        required=True,
        allow_null=False,
        write_only=True,
        display_table=DisplayMode.SUPPRESS,
        label=_("Recipients"),
    )

    message_subject = fields.CharField(write_only=True, label=_("Subject"), display_table=DisplayMode.HIDDEN)
    message_body = MessageBodyField()

    send_on_channels = fields.MultipleChoiceField(
        label=_("Send on channels"),
        allow_empty=False,
        display_table=DisplayMode.SUPPRESS,
        display_form=DisplayMode.FULL,
        choices=[(c.name, c.name) for c in ChannelIdentifier.supported_channels()],
        write_only=True,
    )

    send_notification_sms = fields.BooleanField(
        conditional_visibility=Statement(
            F("send_on_channels").not_includes(lambda: (SmsChannel.name)),
            Operators.AND,
            F("send_on_channels").includes(
                lambda: (
                    list(filter(lambda c: c.name != SmsChannel.name, ChannelIdentifier.supported_channels()))[0].name
                ),
            ),
        ),
        label=_("Send notification SMS"),
        display_table=DisplayMode.HIDDEN,
    )

    created_at = ReadOnlyDateTimeFieldFromTs(
        label=_("Sent"),
        display_form=DisplayMode.HIDDEN,
        read_only=True,
        allow_null=True,
    )

    def get_delivery(self, rec: DjangoProjectBaseNotification):
        req_channels = rec.required_channels.split(",") if rec.required_channels else []
        sent_channels = set(rec.sent_channels.split(",")) if rec.sent_channels else set()
        failed_channels = set(rec.failed_channels.split(",")) if rec.failed_channels else set()
        res = []
        for channel in req_channels:
            if channel in sent_channels:
                res.append(f'{channel} <span style="color: green">\u2714</span>')
            elif channel in failed_channels:
                title_attr = f"title=\"{rec.exceptions if rec.exceptions else ''}\""
                res.append(f'{channel} <span style="color: red" {title_attr}>\u2717</span>')
            else:
                res.append(f"{channel}")
        return ",".join(res)

    def to_representation(self, instance, row_data=None):
        repr = super().to_representation(instance, row_data)
        kw = getattr(self.context.get("view", object()), "kwargs", dict())
        if kw.get("pk") == "new" and kw.get("format") == "componentdef":
            # enable conditional field render in DF
            repr["send_on_channels"] = []
        return repr

    def get_subject(self, obj):
        if not obj or not obj.message:
            return None
        return obj.message.subject

    class Meta:
        model = DjangoProjectBaseNotification
        exclude = (
            "content_entity_context",
            "locale",
            "sent_at",
            "delayed_to",
            "recipients_original_payload_search",
            "exceptions",
            "extra_data",
            "counter",
            "required_channels",
            "sent_channels",
            "failed_channels",
        )
        layout = Layout(
            Row(Column("message_to")),
            Row(Column("message_subject")),
            Row(Column("message_body")),
            Row(Column("send_on_channels")),
            size="large",
        )
        responsive_columns = ResponsiveTableLayouts(
            auto_generate_single_row_layout=True,
            auto_generate_single_column_layout=False,
            layouts=[
                ResponsiveTableLayout(
                    "subject", "recipients_original_payload", "created_at", auto_add_non_listed_columns=False
                ),
                ResponsiveTableLayout(
                    [["subject", "created_at"], "recipients_original_payload"], auto_add_non_listed_columns=False
                ),
                ResponsiveTableLayout("subject", "created_at", auto_add_non_listed_columns=False),
                ResponsiveTableLayout(["subject", "created_at"], auto_add_non_listed_columns=False),
            ],
        )


class MessageToListField(fields.ListField):
    def __init__(self, **kw):
        super().__init__(
            child=fields.CharField(),
            display_table=DisplayMode.SUPPRESS,
            **kw,
        )

    @staticmethod
    def parse(val: list, return_instances=False):
        users = list(filter(lambda i: i and "-" not in i and i.isnumeric(), map(str, val)))
        other_objects = list(filter(lambda i: i and "-" in i, map(str, val)))  # string 'RECORDID-CONTENTTYPEID'
        user_model = get_user_model()
        profile_model = swapper.load_model("django_project_base", "Profile")
        instances = []
        if return_instances:
            instances = list(
                filter(
                    lambda f: f,
                    [user_model.objects.filter(pk=u).first() for u in users],
                )
            )
        for obj in other_objects:
            _data = obj.split("-")
            instance = ContentType.objects.get(pk=_data[1]).model_class().objects.filter(pk=_data[0]).first()
            if not instance:
                continue
            if return_instances:
                instances += [instance]
                continue
            if isinstance(instance, (user_model, profile_model)):
                users += [instance.pk]
                continue
            if items_manager := next(filter(lambda i: "taggeditemthrough" in i, dir(instance)), None):
                for item in getattr(instance, items_manager).all():
                    if cont_object := getattr(item, "content_object", None):
                        if isinstance(cont_object, (user_model, profile_model)):
                            users += [cont_object.userprofile.pk]
                        elif user_related_fields := [
                            f
                            for f in cont_object._meta.fields
                            if isinstance(f, ForeignKey) and f.model in (profile_model, user_model)
                        ]:
                            for field in user_related_fields:
                                users += field.model.objects.filter(**{field.attname: cont_object.pk}).values_list(
                                    get_pk_name(field.model), flat=True
                                )
                        elif related_objects := [
                            item
                            for sub_list in [
                                # Jure: it doesn't matter that it's ContentType, as long as it's an empty queryset
                                [o for o in getattr(cont_object, i.name, ContentType.objects.none()).all()]
                                for i in cont_object._meta.related_objects
                            ]
                            for item in sub_list
                        ]:
                            users += list(
                                map(
                                    lambda g: g.pk,
                                    [
                                        getattr(obj, f.name, object)
                                        for obj in related_objects
                                        for f in obj._meta.fields
                                        if isinstance(f, ForeignKey)
                                        and isinstance(getattr(obj, f.name, object()), (profile_model, user_model))
                                    ],
                                )
                            )
        if return_instances:
            return instances
        return list(set(map(str, users)))

    def get_value(self, dictionary):
        value = super().get_value(dictionary)
        if not value:
            return []
        if value == empty:
            return MessageToListField.parse([])
        if isinstance(value[0], list):
            value = [item for sublist in value for item in sublist]
        return MessageToListField.parse(value)


class NotificationViewset(ModelViewSet):
    authentication_classes = [
        SessionAuthentication,
        BasicAuthentication,
        TokenAuthentication,
    ]
    permission_classes = [IsAuthenticated]
    pagination_class = ModelViewSet.generate_paged_loader(ordering=["-created_at"])

    def get_permissions(self):
        if self.action in ("notification_login", "notification_view"):
            return []
        else:
            return super().get_permissions()

    @extend_schema(exclude=True)
    @action(
        detail=True,
        methods=["GET"],
        url_name="notification-login",
        url_path="info",
        authentication_classes=[],
    )
    def notification_login(self, request, pk=None) -> HttpResponse:
        if not pk:
            raise PermissionDenied
        return render(
            request,
            "notification_login.html",
            dict(
                identifier=pk,
                url=reverse("notification-notification-view"),
                title=swapper.load_model("django_project_base", "Project")
                .objects.get(slug=DjangoProjectBaseNotification.objects.get(pk=pk).project_slug)
                .name,
            ),
        )

    @extend_schema(exclude=True)
    @action(
        methods=["POST"],
        detail=False,
        url_name="notification-view",
        authentication_classes=[],
        url_path="info-view",
    )
    def notification_view(self, request: Request, *args, **kwargs) -> HttpResponse:
        number: Optional[str] = request.data.get("number")
        identifier: Optional[str] = request.data.get("identifier")
        guest_accesses: List[float] = cache.get(identifier, [])
        now: float = time.time()
        if len(list(filter(lambda t: now - t < 60, guest_accesses))) > 4:
            # basic throttling
            raise PermissionDenied()
        guest_accesses.append(now)
        cache.set(identifier, guest_accesses, timeout=120)
        if not number or not identifier or len(number) < 4:
            raise ValidationError()
        notification: DjangoProjectBaseNotification = DjangoProjectBaseNotification.objects.filter(
            pk=identifier
        ).first()
        if not notification:
            raise Http404()

        phone_number_check = (
            get_user_model()
            .objects.filter(pk__in=notification.recipients.split(","))
            .filter(userprofile__phone_number__endswith=number)
            .exists()
        )
        if not phone_number_check:
            raise ValidationError()
        return render(
            request,
            "notification.html",
            dict(message=notification.message.subject + "</br></br></br>" + notification.message.body),
        )

    def filter_queryset_field(self, queryset, field, value):
        if field == "created_at" and value and not value.isnumeric():
            # TODO: search by user defined time range
            value = int(datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ").timestamp())
            return queryset.filter(**{f"{field}__gte": value - 1800, f"{field}__lte": value + 1800})

        if field == "recipients_original_payload" and value:
            return queryset.filter(**{"recipients_original_payload_search__icontains": value})
        return super().filter_queryset_field(queryset, field, value)

    def get_serializer_class(self):
        if self.action in ("create", "update"):

            class NewMessageSerializer(Serializer):
                form_titles = {"edit": _("New message")}

                message_body = NotificationSerializer().fields.fields["message_body"]
                message_subject = NotificationSerializer().fields.fields["message_subject"]
                message_to = MessageToListField(allow_null=False, allow_empty=False)
                send_on_channels = fields.ListField(
                    label=_("Send on channels"),
                    child=fields.CharField(),
                    required=True,
                    display_table=DisplayMode.SUPPRESS,
                    display_form=DisplayMode.SUPPRESS,
                )
                send_notification_sms = fields.BooleanField(default=False, allow_null=False)

            return NewMessageSerializer
        return NotificationSerializer

    def get_queryset(self):
        try:
            return DjangoProjectBaseNotification.objects.filter(
                project_slug=self.request.selected_project_slug
            ).order_by("-created_at")
        except ProjectNotSelectedError as e:
            raise NotFound(e.message)

    def _create_notification(self, serializer):
        host_url = get_host_url(self.request)
        notification = Notification(
            message=DjangoProjectBaseMessage(
                subject=serializer.validated_data["message_subject"],
                body=serializer.validated_data["message_body"],
                footer="",
                content_type=DjangoProjectBaseMessage.HTML,
            ),
            raw_recipents=self.request.data["message_to"],
            project=swapper.load_model("django_project_base", "Project")
            .objects.get(slug=self.request.selected_project_slug)
            .slug,
            recipients=serializer.validated_data["message_to"],
            delay=int(datetime.datetime.now().timestamp()),
            channels=[
                ChannelIdentifier.channel(c, settings=settings, project_slug=None).__class__
                for c in serializer.validated_data["send_on_channels"]
            ],
            persist=True,
            user=self.request.user.pk,
            send_notification_sms=serializer.validated_data["send_notification_sms"],
            host_url=host_url,
        )
        notification.send()

    def update(self, request, *args, **kwargs):
        kwargs.pop("pk", None)
        return super().create(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        kwargs.pop("pk", None)
        return super().create(request, *args, **kwargs)

    def perform_create(self, serializer):
        return self._create_notification(serializer)


class NotificationsLicenseSerializer(LicenseReportSerializer):
    template_context = dict(url_reverse="notification-license")

    def __init__(self, *args, is_filter: bool = False, **kwds):
        super().__init__(*args, is_filter=is_filter, **kwds)
        if (request := self.context.get("request")) and not fields.BooleanField().to_internal_value(
            request.query_params.get("decorate-max-price", "0")
        ):
            self.fields.pop("max_notification_price", None)
        for channel in ChannelIdentifier.supported_channels():
            self.fields[channel.name] = fields.IntegerField(read_only=True, label=_(channel.name))

    usage_report = None
    max_notification_price = fields.FloatField(read_only=True, display_table=DisplayMode.HIDDEN)
    actions = Actions(
        FormButtonAction(btn_type=FormButtonTypes.CANCEL, name="cancel"),
        add_default_crud=False,
        add_default_filter=False,
        add_form_buttons=False,
    )

    class Meta:
        layout = Layout(Row("credit", "used_credit", "remaining_credit"), Row("EMail", "SMS"))


class NotificationsLicenseViewSet(SingleRecordViewSet):
    serializer_class = NotificationsLicenseSerializer

    permission_classes = (IsAuthenticated,)

    def new_object(self):
        license = LogAccessService().report(user=self.request.user)
        usage = 0
        if notifications_usage := next(
            filter(
                lambda u: u.get("item", "") == DjangoProjectBaseNotification._meta.verbose_name,
                license["usage_report"],
            ),
            None,
        ):
            usage = notifications_usage["usage_sum"]
        license["used_credit"] = usage

        license["channels"] = {}
        max_price = 0
        for channel in ChannelIdentifier.supported_channels():
            price = channel.notification_price
            if price > max_price:
                max_price = price
            license[channel.name] = int(max([license["remaining_credit"] / price, 0]))
        if fields.BooleanField().to_internal_value(self.request.query_params.get("decorate-max-price", "0")):
            license["max_notification_price"] = max_price

        return NotificationsLicenseSerializer(license).data

    def create(self, request, *args, **kwargs) -> Response:
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)



import requests

from django.http import FileResponse, HttpResponse
from django.views import static


def documentation_view(request: object, path="", document_root=None) -> FileResponse:
    if not path:
        path = "index.html"
    return static.serve(request=request, path=path, document_root=document_root)


def browser_update_script(request: object) -> HttpResponse:
    # TODO: does this configurability in script name ever come into play?
    try:
        script = request.path.split("js-script")[-1].strip("/")
    except Exception:
        script = ""
    script = script or "update.min.js"
    # TODO: not sure if this proxying is useful at all? I would also suggest cache-ing.
    http_response: HttpResponse = HttpResponse(
        requests.get("https://browser-update.org/%s" % script, verify=False).content.decode(),
        content_type="application/javascript",
    )
    return http_response



from django.conf import settings
from django.db.models.functions import Length
from django.utils import timezone

from django_project_base.notifications.models import DjangoProjectBaseNotification


class DuplicateNotificationMixin(object):
    def handle_similar_notifications(self, notification: DjangoProjectBaseNotification) -> bool:
        message_length: int = len(notification.message.body)
        sent_notifications = (
            DjangoProjectBaseNotification.objects.annotate(message_len=Length("message__body"))
            .filter(
                content_entity_context=notification.content_entity_context,
                created_at__gt=timezone.now().timestamp() - settings.NOTIFICATION_AGGREGATION_TIMEDELTA_SECONDS,
                level=notification.level,
                type=notification.type,
                recipients=notification.recipients,
                required_channels=notification.required_channels,
                message_len__gte=message_length - settings.NOTIFICATION_LENGTH_SIMILARITY_BUFFER_VALUE,
                message_len__lte=message_length + settings.NOTIFICATION_LENGTH_SIMILARITY_BUFFER_VALUE,
                message__subject=notification.message.subject,
                locale=notification.locale,
            )
            .order_by("-created_at")
        )

        if sent_notifications.exists():
            first_same_notification = sent_notifications.first()
            first_same_notification.counter = first_same_notification.counter + 1
            first_same_notification.save(update_fields=["counter"])
            return True
        return False



from pathlib import Path

from django_project_base.settings_parser import parse_settings

DJANGO_PROJECT_BASE_SETTINGS = (
    {
        "name": "DJANGO_PROJECT_BASE_BASE_REQUEST_URL_VARIABLES",
        "default": {
            "project": {
                "value_name": "current_project_slug",
                "url_part": (
                    1,
                    ("project",),
                ),
            },
            "language": {
                "value_name": "current_language",
                "url_part": (
                    1,
                    ("language",),
                ),
            },
        },
    },
    {
        "name": "DJANGO_PROJECT_BASE_SLUG_FIELD_NAME",
        "default": "slug",
    },
    {
        "name": "REST_REGISTRATION",
        "default": {
            "RESET_PASSWORD_VERIFICATION_EMAIL_SENDER": "django_project_base.account.service.reset_password_email_service.send_reset_password_verification_email",  # noqa: E501
            "RESET_PASSWORD_VERIFICATION_URL": "/#reset-user-password/",
            "VERIFICATION_FROM_EMAIL": "",
            "SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL": True,
            "RESET_PASSWORD_VERIFICATION_ENABLED": True,
            "RESET_PASSWORD_SERIALIZER_PASSWORD_CONFIRM": True,
            "SEND_RESET_PASSWORD_LINK_USER_FINDER": "django_project_base.account.service.reset_password_email_service.find_user_by_send_reset_password_link_data",  # noqa: E501
            "REGISTER_VERIFICATION_ENABLED": True,
            "REGISTER_VERIFICATION_EMAIL_SENDER": "django_project_base.account.service.register_user_service.send_register_verification_email_notification",  # noqa: E501
        },
    },
    {"name": "NOTIFICATION_SENDERS", "default": {}},
    {"name": "SYSTEM_EMAIL_SENDER_ID", "default": ""},
    {"name": "SYSTEM_SMS_SENDER_ID", "default": ""},
    {
        "name": "NOTIFICATIONS_EMAIL_PROVIDER",
        "default": "django_project_base.notifications.base.channels.integrations.aws_ses.AwsSes",
    },
    {
        "name": "NOTIFICATIONS_SMS_PROVIDER",
        "default": [
            "django_project_base.notifications.base.channels.integrations.t2.T2",
            "django_project_base.notifications.base.channels.integrations.aws_sns_single_sms.AwsSnsSingleSMS",
            "django_project_base.notifications.base.channels.integrations.nexmo_sms.NexmoSMS",
        ],
    },
    # this settings silences (rest_registration.E013) SEND_RESET_PASSWORD_LINK_SERIALIZER_USE_EMAIL
    # is set but email field is not unique
    # todo: TASK https://taiga.velis.si/project/velis-django-project-admin/us/637?no-milestone=1
    {"name": "SILENCED_SYSTEM_CHECKS", "default": ["rest_registration.E013"]},
    {"name": "CONFIRMATION_CODE_TIMEOUT", "default": 600},
    {"name": "VERIFICATION_FROM_EMAIL", "default": ""},
    {"name": "NOTIFICATION_AGGREGATION_TIMEDELTA_SECONDS", "default": 120},
    {"name": "NOTIFICATION_LENGTH_SIMILARITY_BUFFER_VALUE", "default": 3},
    {
        "name": "LICENSE_ACCESS_USE_CONTENT_TYPE_MODEL",
        "default": "notifications.DjangoProjectBaseNotification",
    },
    {
        "name": "DEFAULT_EMAIL_SENDER_ID_SETTING_NAME",
        "default": "",
    },
    {
        "name": "DEFAULT_SMS_SENDER_ID_SETTING_NAME",
        "default": "",
    },
    {"name": "IS_PHONE_NUMBER_ALLOWED_FUNCTION", "default": ""},
)

USER_CACHE_KEY = "django-user-{id}"
CACHE_IMPERSONATE_USER = "impersonate-user-%d"

PROFILER_LOG_LONG_REQUESTS_COUNT = 50

PROFILE_REVERSE_FULL_NAME_ORDER = False

DELETE_PROFILE_TIMEDELTA = 3

DOCUMENTATION_DIRECTORY: str = str(Path().resolve()) + "/docs/build/"

TEST_USER_ONE_DATA = dict(
    username="miha",
    last_name="Novak",
    first_name="Miha",
    email="user1@user1.si",
    password="mihamiha",
)
TEST_USER_TWO_DATA = dict(
    username="janez",
    last_name="Novak",
    first_name="Janez",
    email="user2@user2.si",
    password="janezjanez",
)


def set_django_project_base_settings():
    parse_settings(DJANGO_PROJECT_BASE_SETTINGS)



from django.utils.crypto import get_random_string

SECRET_KEY = get_random_string(length=64)

# todo: read this setting from env
NOTIFICATION_SEND_PAUSE_SECONDS = 1

# todo: read this setting from env
NOTIFICATION_QUEABLE_HARD_TIME_LIMIT = 180

# todo: read this setting from env
NOTIFICATIONS_QUEUE_VISIBILITY_TIMEOUT = 86400



from typing import Optional, Union

import boto3

from django.conf import Settings
from rest_framework.status import is_success

from django_project_base.notifications.base.channels.channel import Recipient
from django_project_base.notifications.base.channels.integrations.provider_integration import ProviderIntegration
from django_project_base.notifications.models import DeliveryReport, DjangoProjectBaseNotification


class AwsSnsSingleSMS(ProviderIntegration):
    key_id: str
    access_key: str
    region: str

    def __init__(self) -> None:
        super().__init__(settings=object())

    def ensure_credentials(self, settings: Optional[Settings] = None):
        if settings and getattr(settings, "TESTING", False):
            return
        self.key_id = getattr(settings, "NOTIFICATIONS_AWS_SES_ACCESS_KEY_ID", None)
        self.access_key = getattr(settings, "NOTIFICATIONS_AWS_SES_SECRET_ACCESS_KEY", None)
        self.region = getattr(settings, "NOTIFICATIONS_AWS_SES_REGION_NAME", None)
        self.settings = settings
        assert self.key_id, "AWS SES key id required"
        assert self.access_key, "AWS SES key id access key required"
        assert self.region, "AWS SES region required"

    def _get_sms_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return super()._get_sms_message(notification)

    def validate_send(self, response: dict):
        assert response
        assert is_success(response.get("ResponseMetadata", {}).get("HTTPStatusCode", 500))

    def client_send(self, sender: str, recipient: Recipient, msg: str, dlr_id: str):
        if not recipient.phone_number:
            return
        smsattrs = {
            "AWS.SNS.SMS.SenderID": {"DataType": "String", "StringValue": sender.replace(" ", "-")},
            "AWS.SNS.SMS.SMSType": {"DataType": "String", "StringValue": "Promotional"},
        }

        res = (
            boto3.Session(
                aws_access_key_id=self.key_id,
                aws_secret_access_key=self.access_key,
                region_name=self.region,
            )
            .client("sns")
            .publish(PhoneNumber=recipient.phone_number, Message=msg, MessageAttributes=smsattrs)
        )
        self.validate_send(res)

    def get_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return self._get_sms_message(notification)

    def parse_delivery_report(self, dlr: DeliveryReport):
        pass

    @property
    def delivery_report_username_setting_name(self) -> str:
        return "aws-sns-sms-dlr-user"

    @property
    def delivery_report_password_setting_name(self) -> str:
        return "aws-sns-sms-dlr-password"

    def ensure_dlr_user(self, project_slug: str):
        pass

    def enqueue_dlr_request(self, pk: str):
        pass



# Generated by Django 4.2.4 on 2023-09-06 12:13
import swapper
import random
import string
from django.db import migrations


def forwards_func(apps, schema_editor):
    from django_project_base.notifications.base.channels.integrations.t2 import T2

    ProjectModel = swapper.load_model("django_project_base", "Project")
    ProjectModelSettings = swapper.load_model("django_project_base", "ProjectSettings")
    t2 = T2()
    username = t2.delivery_report_username_setting_name
    passwd = t2.delivery_report_password_setting_name
    for project in ProjectModel.objects.all():
        ProjectModelSettings.objects.create(
            name=username,
            project=project,
            description="T2 username for DLR user",
            value=f"t2-{project.slug}-username",
            value_type="char",
        )

        ProjectModelSettings.objects.create(
            name=passwd,
            project=project,
            description="T2 password for DLR user",
            value=f"t2-{project.slug}-password-{''.join(random.choice(string.ascii_letters) for x in range(5))}",
            value_type="char",
        )


def reverse_func(apps, schema_editor):
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0004_alter_djangoprojectbasenotification_options"),
        ("notifications", "0004_alter_djangoprojectbasenotification_options_and_more"),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]



from typing import List

from django.core.management import BaseCommand

from django_project_base.profiling.middleware import ProfileRequest


class PerformanceCommand(BaseCommand):
    def execute(self, *args, **options):
        param_names: list = list(
            filter(
                lambda o: o
                and o
                not in ("pythonpath", "no_color", "force_color", "verbosity", "skip_checks", "traceback", "settings"),
                options,
            )
        )
        command_class_data: List[str] = str(self.__class__).split(".")
        command_name: str = command_class_data[len(command_class_data) - 2]
        name: str = f"manage_command_{command_name}"
        params = ""
        if args:
            params += f"args={args} "
        for param in param_names:
            params += f"{param}={options.get(param)} "

        with ProfileRequest(
            {"REQUEST_METHOD": "MANAGEMENT_COMMAND", "HTTP_HOST": "", "QUERY_STRING": params, "PATH_INFO": name},
            super().execute,
            args,
            options,
        ) as pr:
            return pr.response



from collections.abc import Iterable
from typing import Any


def parse_settings(input_settings: tuple) -> None:
    from django.conf import settings

    for _setting in input_settings:
        setting_name: str = _setting["name"]
        setting_default_value: Any = _setting["default"]
        setting_value: Any = getattr(_setting, _setting["name"], setting_default_value)
        if hasattr(settings, setting_name):
            _setting_existing_val: Any = getattr(settings, setting_name)
            if isinstance(_setting_existing_val, (tuple, list)):
                if isinstance(setting_value, Iterable):
                    _setting_existing_val: list = list(_setting_existing_val)
                    for v in setting_value:
                        _setting_existing_val.append(v)
                    _setting_existing_val: tuple = tuple(sorted(_setting_existing_val, key=_setting_existing_val.index))
                else:
                    _setting_existing_val: list = list(_setting_existing_val)
                    _setting_existing_val.append(setting_value)
                    _setting_existing_val: tuple = tuple(set(_setting_existing_val))
                setattr(settings, setting_name, _setting_existing_val)
            elif isinstance(_setting_existing_val, dict):
                for s_name, s_value in setting_default_value.items():
                    if not _setting_existing_val.get(s_name, None):
                        _setting_existing_val[s_name] = s_value
                setattr(settings, setting_name, _setting_existing_val)
            else:
                pass
        else:
            setattr(settings, setting_name, setting_value)



"""
Project base expects there to be a project and nearly everything else is linked to the project.
Hence this file that offers filtering models and queries by (currently selected) project

There are three utility functions here:

filter_queryset_to_project:
  filters queryset on specified project field or subquery to the given project.
  if no project can be found, empty queryset is returned

filter_queryset_or_model_to_project:
  expands on filter_queryset_to_project and includes smart handling either models or querysets

ProjectFilteringManager:
  a Manager that adds a filter_by_project method to filter model records.
  Easily accessible by setting model objects to it

This file is the basis for viewsets.py ProjectFilteringViewSet
"""

from typing import Optional, Type, TYPE_CHECKING

from django.db.models import Manager, Model, QuerySet

if TYPE_CHECKING:
    from django_project_base.base.models import BaseProject


def filter_queryset_to_project(
    queryset: QuerySet, project_field: str = "project", project: "BaseProject" = None
) -> QuerySet:
    from django_project_base.account.middleware import ProjectNotSelectedError
    from django_project_base.base.middleware import get_current_request, has_current_request

    if not project and has_current_request():
        project = getattr(get_current_request(), "selected_project", None)

    try:
        if project:
            return queryset.filter(**{project_field: project})
    except ProjectNotSelectedError:
        pass
    return queryset.none()


def filter_queryset_or_model_to_project(
    queryset: Optional[QuerySet] = None,
    model: Optional[Type[Model]] = None,
    project_field: str = "project",
    project: Optional["BaseProject"] = None,
) -> QuerySet:
    if model is not None and not hasattr(model, "objects") and callable(model):
        model = model()

    if hasattr(queryset, "filter_by_project"):
        # if the provided queryset is already of the correct manager sort
        objs = queryset
    elif queryset is not None:
        # a queryset has been provided and it needs to be filtered
        return filter_queryset_to_project(queryset, project_field, project)
    elif model is None:
        raise ValueError("ProjectFilteringViewSet requires either queryset or model to be set")
    elif isinstance(model.objects, ProjectFilteringManager):
        # then we check if the provided model has the correct manager
        objs = model.objects
    else:
        # finally we filter the queryset of the provided model
        return filter_queryset_to_project(model.objects, project_field, project)

    return objs.filter_by_project(project)


class ProjectFilteringManager(Manager):
    """
    Manager that allows for filtering model queryset to a specified project records only
    """

    def __init__(self, project_field: str = "project"):
        super().__init__()
        self.project_field = project_field

    def filter_by_project(self, project: "BaseProject" = None):
        return filter_queryset_to_project(self.get_queryset(), self.project_field, project)



from typing import Optional

from django_project_base.caching.cache_queue import CacheQueue
from django_project_base.serialization import CacheLock


class CacheQueueOther(CacheQueue):
    def set_cache(self):
        self.cache = self.django_cache

    # noinspection PyPackageRequirements,PyMethodMayBeStatic
    def get_byte_values(self, values):
        def _get_byte_value(_item):
            if not isinstance(_item, bytes):
                return str(_item).encode('utf-8')
            return _item

        return [_get_byte_value(item) for item in values]

    def rpush(self, *values):
        with CacheLock(self.key):
            cache_list = self.cache.get(self.key, [])
            cache_list.extend(self.get_byte_values(values))
            self.cache.set(self.key, cache_list)
            self.update_timeout()

    def lpush(self, *values):
        with CacheLock(self.key):
            cache_list = self.cache.get(self.key, [])
            cache_list[:0] = reversed(self.get_byte_values(values))
            self.cache.set(self.key, cache_list)
            self.update_timeout()

    def lpop(self, count: Optional[int] = None):
        if not count or count <= 0:
            count = 1
        with CacheLock(self.key):
            cache_list = self.cache.get(self.key, [])
            ret = cache_list[:count]
            cache_list = cache_list[count:]
            self.cache.set(self.key, cache_list)
            self.update_timeout()
        if not ret:
            return None
        if count > 1:
            return ret
        else:
            return ret[0]

    def rpop(self, count: Optional[int] = None):
        if not count or count <= 0:
            count = 1
        with CacheLock(self.key):
            cache_list = self.cache.get(self.key, [])
            ret = cache_list[-count:]
            cache_list = cache_list[:-count]
            self.cache.set(self.key, cache_list)
            self.update_timeout()
        if not ret:
            return None
        if count > 1:
            ret.reverse()
            return ret
        else:
            return ret[0]

    def lrange(self, count=None):
        if count is not None:
            count = count + 1
        return self.cache.get(self.key, [])[0:count]

    def ltrim(self, count=None):
        with CacheLock(self.key):
            cache_list = self.cache.get(self.key, [])
            cache_list = cache_list[count:None]
            self.cache.set(self.key, cache_list)
            self.update_timeout()

    def update_timeout(self):
        self.cache.touch(self.key, self.timeout)



import celery

from django_project_base.profiling.performance_function_decorator import function_profiler


class PerformanceCeleryTask(celery.Task):
    def get_profiler_params(self):
        return dict()

    def __getattribute__(self, name):
        # Get the original attribute
        attr = super().__getattribute__(name)

        # If the attribute is a callable (i.e., a method), wrap it with the custom behavior
        if name == "run" and callable(attr):
            return function_profiler(**self.get_profiler_params())(attr)
        return attr



from rest_framework.routers import DefaultRouter

from django_project_base.licensing.rest.rest import LicenseViewSet

licensing_router = DefaultRouter()

licensing_router.register(
    r"license",
    LicenseViewSet,
    basename="license",
)



from rest_framework.authentication import SessionAuthentication


# noinspection PyRedeclaration
class SessionAuthentication(SessionAuthentication):
    """
    This is needed because original SessionAuthentication doesn't return anything is authenticate_header. And this means
    that if user is not logged in, status of response would be 403 and not 401.
    If authenticate_header returns some string, status of response will be 401.

    This is only relevant for authentication class that if listed first in
    REST_FRAMEWORK / DEFAULT_AUTHENTICATION_CLASSES setting.
    Also see code in rest_framework.views.APIView.handle_exception and inside this function see call of function
    self.get_authenticate_header

    If you use this SessionAuthentication class instad of original one
    (rest_framework.authentication.SessionAuthentication) you must also set
    REST_REGISTRATION["LOGIN_AUTHENTICATE_SESSION"] to True in settings.py.
    Without it, user won't be logged in.
    See code in rest_registration.api.views.login.LoginView.post. Here is call to perform_login function. Inside that,
    there is call of should_authenticate_session() that returns True if there is original SessionAuthentication class
    listed in REST_FRAMEWORK / DEFAULT_AUTHENTICATION_CLASSES setting. If you are using inherited one,
    it will return False by default. Except you set REST_REGISTRATION["LOGIN_AUTHENTICATE_SESSION"] setting to True in
    settings.py.
    """

    def authenticate_header(self, request):
        return "sessionid"



from gettext import gettext

import swapper

from django.shortcuts import get_object_or_404

from django_project_base.base.event import ProjectSettingConfirmedEvent
from django_project_base.notifications.email_notification import SystemEMailNotification
from django_project_base.notifications.models import DjangoProjectBaseMessage
from django_project_base.profiling.performance_base_command import PerformanceCommand


class Command(PerformanceCommand):
    help = "Confirms project setting. Example:  python manage.py confirm_setting 2 email-sender-id"

    def add_arguments(self, parser) -> None:
        parser.add_argument("project-id", type=str, help="Project identifier")
        parser.add_argument("setting-name", type=str, help="Setting name")

    def handle(self, *args, **options):
        project = get_object_or_404(swapper.load_model("django_project_base", "Project"), pk=str(options["project-id"]))
        setting = get_object_or_404(
            swapper.load_model("django_project_base", "ProjectSettings"),
            project=project,
            name=str(options["setting-name"]),
        )
        ProjectSettingConfirmedEvent(user=None).trigger(payload=setting)
        SystemEMailNotification(
            message=DjangoProjectBaseMessage(
                subject=f"{gettext('Project setting confirmed')}",
                body=f"{gettext('Setting')} {setting.name} {gettext('in project')} "
                f"{project.name} {gettext('has been confirmed and is now active.')}",
                footer="",
                content_type=DjangoProjectBaseMessage.PLAIN_TEXT,
            ),
            recipients=[],  # TODO: find project owner and send, when project owner is found use EmailNotification
            user=None,  # TODO: find project owner and send -> add .send()
        )
        return "ok"



from django.core.cache import cache
from hijack.helpers import hijack_ended, hijack_started

from django_project_base.settings import CACHE_IMPERSONATE_USER


def hijack_set_is_hijacked(sender, **kwargs):
    cache.set(CACHE_IMPERSONATE_USER % kwargs.get("hijacked").id, True)


def hijack_delete_is_hijacked(sender, **kwargs):
    cache.delete(CACHE_IMPERSONATE_USER % kwargs.get("hijacked").id)


hijack_started.connect(hijack_set_is_hijacked)
hijack_ended.connect(hijack_delete_is_hijacked)



from dynamicforms.routers import DFRouter

from django_project_base.account.rest.account import (
    AdminAddUserViewSet,
    ChangePasswordViewSet,
    LogoutViewSet,
    RegisterViewSet,
    SocialAuthProvidersViewSet,
    VerifyRegistrationViewSet,
)
from django_project_base.account.rest.impersonate import ImpersonateUserViewset
from django_project_base.account.rest.invite import ProjectUserInviteViewSet
from django_project_base.account.rest.login import LoginViewset
from django_project_base.account.rest.profile import ProfileViewSet, ProjectsProfileSearchViewSet
from django_project_base.account.rest.profile_merge import ProfileMergeViewSet
from django_project_base.account.rest.reset_password import (
    InvalidatePasswordAdminViewSet,
    ResetPasswordViewSet,
    SendResetPasswordLinkViewSet,
)

accounts_router = DFRouter(trailing_slash=True)

accounts_router.register(r"", SocialAuthProvidersViewSet, basename="account")
accounts_router.register(r"change-password", ChangePasswordViewSet, basename="change-password")
accounts_router.register(
    r"admin-invalidate-password", InvalidatePasswordAdminViewSet, basename="admin-invalidate-password"
)
accounts_router.register(r"admin-add-user", AdminAddUserViewSet, basename="admin-add-user")
accounts_router.register(r"", ResetPasswordViewSet, basename="account-reset-pass")
accounts_router.register(r"", RegisterViewSet, basename="account-register")
accounts_router.register(r"", SendResetPasswordLinkViewSet, basename="account-send-reset-pwd-link")
accounts_router.register(r"", VerifyRegistrationViewSet, basename="account-verify-registration")

profile_router = DFRouter(trailing_slash=False)
profile_router.register(r"profile", ProfileViewSet, basename="profile-base-project")
profile_router.register(r"profile-merge", ProfileMergeViewSet, basename="profile-merge-base-project")
profile_router.register_single_record(r"impersonate", ImpersonateUserViewset, basename="profile-base-impersonate-user")
profile_router.register_single_record(r"login", LoginViewset, basename="profile-base-login")
# Following line is for backwards compatibility. PBXs calls login with trailing slash
profile_router.register_single_record(r"login/", LoginViewset, basename="profile-base-login2")
profile_router.register(r"", LogoutViewSet, basename="account-logout")
profile_router.register(r"project-user-invite", ProjectUserInviteViewSet, basename="project-user-invite")
profile_router.register(r"projects-profile-search", ProjectsProfileSearchViewSet, basename="projects-profile-search")



import datetime
import uuid

import swapper

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.validators import int_list_validator
from django.db import models, OperationalError, ProgrammingError
from django.db.models import SET_NULL, Value
from django.db.models.functions import Concat
from django.utils.translation import gettext_lazy as _

from django_project_base.notifications.base.enums import NotificationLevel, NotificationType
from django_project_base.notifications.notification_queryset import NotificationQuerySet
from django_project_base.utils import get_pk_name, IntDescribedEnum


class AbstractDjangoProjectBaseMessage(models.Model):
    PLAIN_TEXT = "text/plain"
    HTML = "text/html"

    CONTENT_TYPE_CHOICES = ((PLAIN_TEXT, "Plain text"), (HTML, "Html"))

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, verbose_name=_("Id"))
    subject = models.TextField(null=True, blank=True, verbose_name=_("Subject"))
    body = models.TextField(verbose_name=_("Body"))
    footer = models.TextField(null=True, blank=True, verbose_name=_("Footer"))
    content_type = models.CharField(null=False, choices=CONTENT_TYPE_CHOICES, default=PLAIN_TEXT, max_length=64)

    class Meta:
        abstract = True
        verbose_name = "Message"


class DjangoProjectBaseMessage(AbstractDjangoProjectBaseMessage):
    pass


def integer_ts():
    return int(datetime.datetime.now().timestamp())


class AbstractDjangoProjectBaseNotification(models.Model):
    locale = models.CharField(null=True, blank=True, max_length=8, verbose_name=_("Locale"))  # language
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, verbose_name=_("Id"))
    level = models.CharField(
        null=False,
        blank=False,
        max_length=16,
        choices=[(i.value, i.name.lower().title()) for i in NotificationLevel],
        verbose_name=_("Level"),
    )
    required_channels = models.CharField(null=True, blank=True, max_length=32, verbose_name=_("Required channels"))
    sent_channels = models.CharField(null=True, blank=True, max_length=32, verbose_name=_("Sent channels"))
    failed_channels = models.CharField(null=True, blank=True, max_length=32, verbose_name=_("Failed channels"))
    created_at = models.BigIntegerField(default=integer_ts, editable=False, verbose_name=_("Created at"))
    sent_at = models.BigIntegerField(null=True, blank=True, verbose_name=_("Sent at"))
    delayed_to = models.BigIntegerField(null=True, blank=True, verbose_name=_("Delayed to"))
    type = models.CharField(
        null=False,
        blank=False,
        max_length=16,
        choices=[(i.value, i.name.lower().title()) for i in NotificationType],
        default=NotificationType.STANDARD.value,
        verbose_name=_("Type"),
    )
    recipients = models.CharField(blank=False, null=True, max_length=2048, validators=[int_list_validator])
    message = models.OneToOneField(
        DjangoProjectBaseMessage,
        on_delete=SET_NULL,
        null=True,
        verbose_name=_("Message"),
    )
    exceptions = models.TextField(null=True)
    content_entity_context = models.TextField()
    counter = models.SmallIntegerField(default=1)
    recipients_original_payload = models.CharField(blank=False, null=False, max_length=2048)
    recipients_original_payload_search = models.TextField(blank=False, null=True, db_index=True)
    project_slug = models.CharField(null=True, blank=True, max_length=1024)
    send_notification_sms = models.BooleanField(null=False, blank=False, default=False)
    send_notification_sms_text = models.TextField(blank=False, null=True)
    extra_data = models.JSONField(null=True, blank=False)

    class Meta:
        abstract = True
        verbose_name = "Notification"


class DjangoProjectBaseNotification(AbstractDjangoProjectBaseNotification):
    objects = NotificationQuerySet.as_manager()

    _recipients_list = []
    _user = None
    _sender = {}
    _email_list = []
    _email_fallback = False

    def _get_recipients(self):
        return self._recipients_list

    def _set_recipents(self, val):
        self._recipients_list = val

    def _get_user(self):
        return self._user

    def _set_user(self, val):
        self._user = val

    def _get_sender(self):
        return self._sender

    def _set_sender(self, val):
        self._sender = val

    def _get_email_list(self):
        return self._email_list

    def _set_email_list(self, val):
        self._email_list = val

    def _get_email_fallback(self):
        return self._email_fallback

    def _set_email_fallback(self, val):
        self._email_fallback = val

    recipients_list = property(_get_recipients, _set_recipents)

    user = property(_get_user, _set_user)

    sender = property(_get_sender, _set_sender)

    email_list = property(_get_email_list, _set_email_list)

    email_fallback = property(_get_email_fallback, _set_email_fallback)

    @property
    def recipients_raw(self):
        ids = [int(recipients_id) for recipients_id in self.recipients.split(",")] if self.recipients else []
        return swapper.load_model("django_project_base", "Profile").objects.filter(id__in=ids)


class SearchItemObject:
    label = ""

    def __init__(self, dict_val):
        for key, value in dict_val.items():
            setattr(self, key, value)
        setattr(self, "pk", dict_val["ido"])
        setattr(self, "id", dict_val["ido"])

    def __str__(self) -> str:
        return self.label


class SearchItemsManager(models.Manager):
    def get_queryset(self, **kwargs):
        tag_model = swapper.load_model("django_project_base", "Tag")

        slug = ""

        if not kwargs.get("request"):
            try:
                from django_project_base.base.middleware import get_current_request

                request = get_current_request()
                if request and getattr(request, "selected_project_slug", None):
                    slug = request.selected_project_slug
            except Exception:
                pass
        else:
            slug = getattr(kwargs["request"], "selected_project_slug", "")

        if not slug:
            return []

        try:
            tag_model_content_type_id = ContentType.objects.get_for_model(tag_model).pk
            user_model_content_type_id = ContentType.objects.get_for_model(get_user_model()).pk
            user_model = get_user_model()
            qs = []
            qs += [
                SearchItemObject(o)
                for o in user_model.objects.filter(userprofile__projects__project__slug=slug)
                .distinct()
                .annotate(  # qs users
                    ido=Concat(
                        get_pk_name(user_model),
                        Value("-"),
                        Value(user_model_content_type_id),
                        output_field=models.CharField(),
                    )
                )
                .exclude(is_active=False)
                .extra(
                    select={
                        "object_id": f"{user_model._meta.db_table}.{get_pk_name(user_model)}",
                        "label": "first_name || ' ' || last_name",
                        "content_type_id": user_model_content_type_id,
                    }
                )
                .values("object_id", "label", "content_type_id", "ido")
            ]
            tgs = [
                SearchItemObject(o)
                for o in tag_model.objects.filter(project__slug=slug)
                .distinct()
                .annotate(  # qs tags
                    ido=Concat(
                        get_pk_name(tag_model),
                        Value("-"),
                        Value(tag_model_content_type_id),
                        output_field=models.CharField(),
                    )
                )
                .extra(
                    select={
                        "object_id": f"{tag_model._meta.db_table}.{get_pk_name(tag_model)}",
                        "label": f"{tag_model._meta.db_table}.name",
                        "content_type_id": tag_model_content_type_id,
                    }
                )
                .values("object_id", "label", "content_type_id", "ido")
            ]
            qs += tgs
            return qs
        except OperationalError:
            return []
        except ProgrammingError:
            return []


class SearchItems(models.Model):
    objects = SearchItemsManager()

    content_type_id = models.PositiveIntegerField()
    object_id = models.PositiveBigIntegerField()
    name = models.CharField(max_length=256)
    id = models.PositiveBigIntegerField(primary_key=True)

    class Meta:
        abstract = False
        managed = False
        db_table = "search-items-table"


class DeliveryReport(models.Model):
    class Status(IntDescribedEnum):
        DELIVERED = 1, _("Delivered")
        NOT_DELIVERED = 2, _("Not Delivered")
        PENDING_DELIVERY = 3, _("Pending delivery")

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, verbose_name=_("Id"), db_index=True)
    notification = models.ForeignKey(DjangoProjectBaseNotification, on_delete=models.CASCADE)
    user_id = models.CharField(max_length=1024, db_index=True, null=False, blank=False)
    channel = models.CharField(max_length=1024, null=False, blank=False)
    provider = models.CharField(max_length=1024, db_index=True, null=False, blank=False)
    payload = models.JSONField(null=True, blank=False)
    status = models.IntegerField(
        default=Status.PENDING_DELIVERY, choices=Status.get_choices_tuple(), db_index=True, null=False
    )
    auxiliary_notification = models.UUIDField(verbose_name=_("Auxiliary notification"), null=True, blank=False)



"""Get enabled social auth providers configuration."""
import importlib

from collections import namedtuple
from functools import reduce
from typing import Iterable, List

from django.conf import settings

from django_project_base.constants import ACCOUNT_URL_PREFIX

SocialProviderItem = namedtuple("SocialProviderItem", ("name", "title", "url"))


def get_social_providers() -> List[SocialProviderItem]:
    config: List[SocialProviderItem] = []

    authentication_backends: Iterable = filter(lambda b: "social_core" in b, settings.AUTHENTICATION_BACKENDS)
    existing_settings: list = list(
        map(lambda e: e.lower(), filter(lambda s: s.lower().startswith("social_auth_"), dir(settings)))
    )
    for auth_bckend in authentication_backends:
        # first, we import the auth backend
        auth_backend = importlib.import_module(".".join(auth_bckend.split(".")[:3]))
        # next we go into the backend's implementation module
        auth_module = reduce(lambda x, attr_name: getattr(x, attr_name), auth_bckend.split(".")[3:], auth_backend)
        # and finally get the backend's module name
        name: str = auth_module.name

        search_query: str = next(iter(name.split("-"))).lower()
        search_results: list = list(
            filter(lambda d: search_query in d and (d.endswith("_key") or d.endswith("_secret")), existing_settings)
        )
        if search_results:
            config.append(
                SocialProviderItem(name, search_query.title(), "/%s/social/login/%s/" % (ACCOUNT_URL_PREFIX, name))
            )

    return config



import re

from typing import Optional

import swapper

from django.contrib.auth import get_user_model, login, update_session_auth_hash
from django.contrib.auth.models import AnonymousUser
from django.core.cache import cache
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.validators import validate_email
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view, OpenApiParameter, OpenApiResponse, OpenApiTypes
from dynamicforms import fields as df_fields, serializers as df_serializers, viewsets as df_viewsets
from dynamicforms.action import Actions
from rest_framework import fields, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.permissions import AllowAny, IsAdminUser, IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer
from rest_registration.api.views import change_password, logout, register
from rest_registration.settings import registration_settings
from social_django.models import UserSocialAuth

from django_project_base.account.rest.reset_password import ResetPasswordSerializer
from django_project_base.account.social_auth.providers import get_social_providers


def get_hijacker(request: Request) -> Optional:
    session = getattr(getattr(getattr(request, "_request", object()), "session", object()), "_session", dict())
    if (
        session.get("is_hijacked_user", False)
        and session.get("hijack_history", False)  # noqa: W503
        and request.user  # noqa: W503
        and request.user != AnonymousUser  # noqa: W503
    ):
        if user_pk := next(iter(session["hijack_history"]), None):
            if hijacker := get_user_model().objects.filter(pk=user_pk).first():
                return hijacker
    return None


class SocialAuthSerializer(ModelSerializer):
    class Meta:
        model = UserSocialAuth
        exclude = ("extra_data", "uid")


class SocialAuthProvidersViewSet(viewsets.ViewSet):
    def get_permissions(self):
        if self.action == "social_auth_providers":
            return []
        elif self.action == "social_auth_providers_user":
            return [IsAuthenticated()]
        else:
            return super().get_permissions()

    @extend_schema(
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
        }
    )
    @action(
        detail=False,
        methods=["get"],
        url_path="social-auth-providers",
        url_name="social-auth-providers",
        authentication_classes=[],
    )
    def social_auth_providers(self, request: Request) -> Response:
        return Response(map(lambda item: item._asdict(), get_social_providers()))

    @extend_schema(
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK", response=SocialAuthSerializer),
        }
    )
    @action(
        detail=False,
        methods=["get"],
        url_path="social-auth-providers-user",
        url_name="social-auth-providers-user",
    )
    def social_auth_providers_user(self, request: Request) -> Response:
        return Response(SocialAuthSerializer(UserSocialAuth.objects.filter(user=self.request.user), many=True).data)


class LogoutSerializer(serializers.Serializer):
    revoke_token = fields.BooleanField(required=False)


class LogoutViewSet(viewsets.ViewSet):
    serializer_class = LogoutSerializer

    def get_permissions(self):
        if self.action == "logout":
            return [IsAuthenticated()]
        else:
            return super().get_permissions()

    def initialize_request(self, request, *args, **kwargs):
        request = super().initialize_request(request, *args, **kwargs)
        request._dont_enforce_csrf_checks = True
        return request

    @extend_schema(
        description="Logs out the user. Returns an error if the user is not authenticated. "
        "If revoke_token is provided, revokes the given token for a given user. If the token is not  "
        "provided, revoke all tokens for given user. ",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not authorised"),
        },
    )
    @action(detail=False, methods=["post"], url_name="logout")
    def logout(self, request: Request) -> Response:
        request._request._dont_enforce_csrf_checks = True
        return logout(request._request)


class ChangePasswordSerializer(df_serializers.Serializer):
    template_context = dict(
        url_reverse="change-password",
        dialog_classes="modal-lg",
        dialog_header_classes="bg-info",
    )
    form_titles = {
        "table": "",
        "new": _("Change password"),
        "edit": "",
    }

    old_password = df_fields.CharField(label=_("Current password"), required=True, password_field=True)
    password = df_fields.CharField(label=_("New password"), required=True, password_field=True)
    password_confirm = df_fields.CharField(label=_("Confirm new password"), required=True, password_field=True)

    actions = Actions(add_form_buttons=True)


@extend_schema_view(
    retrieve=extend_schema(parameters=[OpenApiParameter("id", OpenApiTypes.STR, OpenApiParameter.PATH, enum=["new"])]),
)
class ChangePasswordViewSet(df_viewsets.SingleRecordViewSet):
    serializer_class = ChangePasswordSerializer
    permission_classes = [AllowAny]

    def initialize_request(self, request, *args, **kwargs):
        request = super().initialize_request(request, *args, **kwargs)
        # We need to set following flag (which is used while testing), because otherwise CSRF middleware
        # (django/middleware/csrf.py -> process_view()) would execute request.POST, which would cause
        # "django.http.request.RawPostDataException: You cannot access body after reading from request's data stream"
        # when request will be initialized and authenticated later for rest_registration.api.views.change_password
        #
        # It is quite an ugly hack. But I cant find other way around. And security checks are anyway made by
        # rest_registration.api.views.
        request._dont_enforce_csrf_checks = True
        return request

    def new_object(self):
        return dict(old_password="", password="", password_confirm="")

    @extend_schema(
        description="Change the user password.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    def create(self, request: Request) -> Response:
        def handle_password_changed(request):
            update_session_auth_hash(request, request.user)
            profile_obj = getattr(
                request.user,
                swapper.load_model("django_project_base", "Profile")._meta.model_name,
            )
            profile_obj.password_invalid = False
            profile_obj.save(update_fields=["password_invalid"])

        if get_hijacker(request):

            class SuperUserChangePasswordSerializer(ChangePasswordSerializer):
                old_password = None

            serializer = SuperUserChangePasswordSerializer(
                data=request.data,
                context={"request": request},
            )
            serializer.is_valid(raise_exception=True)
            user = request.user
            user.set_password(serializer.validated_data["password"])
            user.save()
            handle_password_changed(request)
            return Response()

        response = change_password(request._request)
        if response.status_code == status.HTTP_200_OK:
            handle_password_changed(request)
        return response


class VerifyRegistrationViewSet(viewsets.ViewSet):
    def get_permissions(self):
        if self.action == "verify_registration":
            return []
        else:
            return super().get_permissions()

    @extend_schema(
        description="Verify registration with signature. The endpoint will generate an e-mail with a confirmation URL "
        "which the user should GET to confirm the account.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK", response=ResetPasswordSerializer),
            status.HTTP_400_BAD_REQUEST: OpenApiResponse(description="Bad request"),
        },
    )
    @action(
        detail=False,
        methods=["post"],
        url_path="verify-registration",
        url_name="verify-registration",
    )
    def verify_registration(self, request: Request) -> Response:
        if (
            (flow_id := request.COOKIES.get("register-flow"))
            and (code := cache.get(flow_id))
            and (req_code := request.data.get("code"))
            and code == req_code
            and len(code)
            and len(req_code)
            and (user := cache.get(code))
        ):
            user.is_active = True
            user.save(update_fields=["is_active"])
            login(request, user, backend="django_project_base.base.auth_backends.UsersCachingBackend")
            response = Response()
            response.delete_cookie("register-flow")
            return response
        raise ValidationError(dict(code=[_("Code invalid")]))

    @action(
        detail=False,
        methods=["post"],
        url_path="verify-registration-email-change",
        url_name="verify-registration-email-change",
    )
    def verify_registration_change_email(self, request: Request) -> Response:
        if (
            (flow_id := request.COOKIES.get("register-flow"))
            and (code := cache.get(flow_id))
            and len(code)
            and (user := cache.get(code))
        ):
            email = request.data.get("email")
            if not email:
                raise ValidationError(dict(email=[_("Email invalid")]))
            try:
                validate_email(email)
            except DjangoValidationError:
                raise ValidationError(dict(email=[_("Email invalid")]))
            user.email = email
            user.save(update_fields=["email"])
            if registration_settings.REGISTER_VERIFICATION_ENABLED:
                registration_settings.REGISTER_VERIFICATION_EMAIL_SENDER(request=request, user=user)
            return Response()
        raise PermissionDenied


class AbstractRegisterSerializer(df_serializers.Serializer):
    username = df_fields.CharField(required=True)
    email = df_fields.CharField(required=True)
    password = df_fields.CharField(required=True)
    password_confirm = df_fields.CharField(required=True)
    first_name = df_fields.CharField(required=False)
    last_name = df_fields.CharField(required=False)

    class Meta:
        abstract = True


class RegisterSerializer(AbstractRegisterSerializer):
    pass


class RegisterReturnSerializer(serializers.Serializer):
    id = fields.IntegerField()
    username = fields.CharField()
    email = fields.CharField()
    first_name = fields.CharField()
    last_name = fields.CharField()


@extend_schema(
    description="Register new user. The endpoint will generate an e-mail with a confirmation URL which the newly "
    "registered user should GET to activate the newly created account.",
    responses={
        status.HTTP_200_OK: OpenApiResponse(description="OK", response=RegisterReturnSerializer),
        status.HTTP_400_BAD_REQUEST: OpenApiResponse(
            description=re.sub(
                r" +",
                " ",
                """
                   There was a problem with the data. Response contains a JSON object showing validation errors per
                   field. JSON object keys are field names with issues, containing a list of issues and a
                   "non_field_errors" field containing errors that do not apply to a single field. Example: {
                     "password": [
                       "The password is too similar to the username.",
                       "This password is too short. It must contain at least 8 characters."
                     ],
                     "password_confirm": [
                       "Passwords don't match"
                     ],
                     "non_field_errors": []
                   }""".replace("\n", " "),
            )
        ),
    },
)
class RegisterViewSet(viewsets.ViewSet):
    serializer_class = RegisterSerializer

    @action(detail=False, methods=["post"], url_path="register", url_name="register")
    def register(self, request: Request) -> Response:
        return register(request._request)


class AdminAddUserSerializer(AbstractRegisterSerializer):
    template_context = dict(
        url_reverse="admin-add-user",
        dialog_classes="modal-lg",
        dialog_header_classes="bg-info",
    )
    form_titles = {
        "table": "",
        "new": _("Add new user"),
        "edit": "",
    }

    actions = Actions()


@extend_schema_view(
    retrieve=extend_schema(parameters=[OpenApiParameter("id", OpenApiTypes.STR, OpenApiParameter.PATH, enum=["new"])]),
)
class AdminAddUserViewSet(df_viewsets.SingleRecordViewSet):
    serializer_class = AdminAddUserSerializer

    permission_classes = (
        IsAuthenticated,
        IsAdminUser,
    )  # TODO: permission should be based on project role

    def initialize_request(self, request, *args, **kwargs):
        request = super().initialize_request(request, *args, **kwargs)
        request.csrf_processing_done = True
        return request

    def new_object(self):
        return dict(
            username="",
            email="",
            password="",
            password_confirm="",
            first_name="",
            last_name="",
        )



from django.shortcuts import get_object_or_404

from django_project_base.notifications.base.notification import Notification
from django_project_base.notifications.models import DjangoProjectBaseNotification
from django_project_base.profiling.performance_base_command import PerformanceCommand


class Command(PerformanceCommand):
    help = 'Resends notification. Example:  python manage.py resend_notification "178a46c2-2aa3-4a33-bad6-9af2d76f6891" 2'  # noqa: E501

    def add_arguments(self, parser) -> None:
        parser.add_argument("notification", type=str, help="Notification identifier (uuid string).")
        parser.add_argument("user", type=str, help="User identifier (user sending notification).", nargs="?")

    def handle(self, *args, **options):
        Notification.resend(
            get_object_or_404(DjangoProjectBaseNotification, pk=str(options["notification"])), str(options["user"])
        )
        return "ok"



from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from django_project_base.utils import IntDescribedEnum


class LicenseAccessUse(models.Model):
    class UseType(IntDescribedEnum):
        USE = 4, _("Used access")
        ADMIN_USE = 10, _("Used access by admin")

    date = models.DateTimeField(verbose_name=_("Date"), default=now, null=False, blank=False)
    type = models.IntegerField(verbose_name=_("Type"), choices=UseType.get_choices_tuple(), default=UseType.USE)

    user_id = models.CharField(max_length=1024, db_index=True)

    content_type_object_id = models.CharField(max_length=1024)

    content_type = models.ForeignKey(ContentType, on_delete=models.DO_NOTHING)

    comment = models.JSONField(verbose_name=_("Comment"), null=True, blank=False)

    amount = models.FloatField(default=0)



from django.urls import include, path

from django_project_base.account.router import accounts_router, profile_router

urlpatterns = [
    path("", include(accounts_router.urls)),
    path("", include(profile_router.urls)),
]



import functools

from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import BasePermission as DRFBasePermission
from rest_framework.settings import api_settings


def check_permission(permission: str):
    def check(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            request = args[1]
            # view_set = args[0]

            def make_check_permission(permission, request) -> bool:
                # user = request.user
                # project = request.selected_project
                return True

            if make_check_permission(permission, request):
                return func(*args, **kwargs)

            raise PermissionDenied

        return wrap

    return check


class BasePermissions(DRFBasePermission):
    def has_permission(self, request, view):
        default_permission_classes = getattr(api_settings, "DEFAULT_PERMISSION_CLASSES", [])
        return all(
            [permission_class().has_permission(request, view) for permission_class in default_permission_classes]
        )



# Generated by Django 4.2.4 on 2023-11-08 06:58
from django.db import migrations, connection
from django.conf import settings


def forwards_func(apps, schema_editor):
    table = "notifications_djangoprojectbasenotification"
    for name, sql in {
        "recipients_original_payload": f"ALTER TABLE {table} ADD COLUMN %s character varying(2048) NOT NULL DEFAULT ''",
        "recipients_original_payload_search": f"ALTER TABLE {table} ADD COLUMN %s text",
    }.items():
        try:
            with connection.cursor() as cursor:
                if "sqlite" in settings.DATABASES["default"]["ENGINE"].split(".")[-1]:
                    cursor.execute(f"SELECT COUNT(*) AS CNTREC FROM pragma_table_info('{table}') WHERE name='{name}';")
                else:
                    cursor.execute(
                        f"select exists (select 1 from information_schema.columns where table_name ='{table}' and column_name='{name}');"
                    )
                row = cursor.fetchone()
                if row and not next(iter(row), None):
                    cursor.execute(sql % name)
        except Exception as e:
            pass


def reverse_func(apps, schema_editor):
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0008_deliveryreport_auxiliary_notification"),
    ]

    operations = [migrations.RunPython(forwards_func, reverse_func)]



import swapper

from django.contrib.auth.models import Group, Permission


class BaseRole(Group):
    # Right now our role does not do anything more than Django's
    class Meta:
        abstract = True


class Role(BaseRole):
    # Right now our role does not do anything more than Django's
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "Role")


class BasePermission(Permission):
    # Right now our permission does not do anything more than Django's
    class Meta:
        abstract = True


class Permission(BasePermission):
    # Right now our permission does not do anything more than Django's
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "Permission")



from typing import Iterable

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view, OpenApiResponse
from dynamicforms import fields, serializers, viewsets
from dynamicforms.action import Actions, FormButtonAction, FormButtonTypes
from dynamicforms.mixins import DisplayMode
from dynamicforms.template_render.layout import Layout, Row
from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_registration.api.views import login

from django_project_base.account.social_auth.providers import get_social_providers, SocialProviderItem
from django_project_base.base.event import UserLoginEvent


class LoginSerializer(serializers.Serializer):
    form_titles = {"edit": _("Login")}
    actions = Actions(
        FormButtonAction(btn_type=FormButtonTypes.CANCEL, name="cancel", label=_("Back to site")),
        FormButtonAction(btn_type=FormButtonTypes.SUBMIT, name="submit", label=_("Login")),
        add_form_buttons=False,
    )

    template_context = dict(url_reverse="profile-base-login", url_reverse_kwargs=None)
    login = fields.CharField(required=True)
    password = fields.CharField(required=True, password_field=True)
    return_type = fields.ChoiceField(choices=["json", "cookie"], required=False, display_form=DisplayMode.HIDDEN)
    social_auth_providers = fields.SerializerMethodField()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields.update({"return-type": self.fields.pop("return_type")})

    def get_dialog_def(self):
        res = super().get_dialog_def()
        layout = Layout(Row("login", "password", "return-type", "social_auth_providers", "df_control_data"), columns=4)
        res["inline"] = layout.as_component_def(self)
        del res["inline"]["fields"]  # fields are already on this level
        return res

    def get_social_auth_providers(self, unused) -> Iterable[SocialProviderItem]:
        return map(lambda item: item._asdict(), get_social_providers())


@extend_schema_view(
    retrieve=extend_schema(
        description="Retrieves dialog definition for login dialog",
    ),
    create=extend_schema(
        description="<p>Logs in the user with given username and password. </p>"
        "<p>We support two methods of maintaining session information for your client: cookie-based and "
        "header-based.</p>"
        "<p>When you perform the account/login function, you can choose whether the function should return "
        'a session cookie or JSON with session id. Add parameter "return-type" with value "json" to login '
        'function parameters. This will return "sessionid" parameter in returned json instead of cookie. '
        "There is no CSRF when session is passed by the authorization header.</p>"
        "<p>If you choose the cookie, you will then need to supply the cookie(s) to all subsequent "
        "requests. Likewise, if you opt for session id as a variable, you will have to provide "
        "Authorization header to all subsequent requests.</p>"
        "<p>The default uses cookies as those also add a CSRF cookie providing a bit more security. "
        "Use of JSON / header should only be preferred for clients without support for cookies, such as "
        "background maintenance / data exchange scripts.</p>"
        "<p>When using the Authorisation header, use returned session api as token with token type "
        '"sessionid" and returned sessionid as credentials. Authorization: sessionid credentials</p>',
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_400_BAD_REQUEST: OpenApiResponse(
                description="Bad request. Missing either one of parameters or wrong login or password."
            ),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Authentication credentials were not provided."),
        },
    ),
)
class LoginViewset(viewsets.SingleRecordViewSet):
    serializer_class = LoginSerializer
    permission_classes = [AllowAny]

    def initialize_request(self, request, *args, **kwargs):
        # We need to lie to the CSRF middleware that CSRF processing is done, because otherwise CSRF middleware
        # (django/middleware/csrf.py -> process_view()) would execute request.POST, which would parse request data as
        # empty - because it is not encoded with form url encoding, but with JSON.
        #
        # As a result, DRF then doesn't get the JSON payload and can't process the request
        #
        # It is quite an ugly hack. But I cant find other way around. And security checks are anyway made by
        # rest_registration.api.views.
        request = super().initialize_request(request, *args, **kwargs)
        request.csrf_processing_done = True
        return request

    def new_object(self):
        return dict(login="", password="", return_type="cookie")

    def create(self, request: Request, *args, **kwargs) -> Response:
        response = login(request._request)
        if response.renderer_context["request"].data.get("return-type", None) == "json":
            response.data.update({"sessionid": request.session.session_key})
            response.returntype = "json"
        if (
            response.status_code == status.HTTP_200_OK
            and (user_id := request.session.get("_auth_user_id"))
            and (user := get_user_model().objects.filter(pk=user_id).first())
        ):
            UserLoginEvent(user=user).trigger(payload=request)
        response.delete_cookie("register-flow")
        return response



NOTIFICATIONS_SETTINGS = (
    {
        "name": "MAINTENANCE_NOTIFICATIONS_CACHE_TIMEOUT",
        "default": 30,
        "description": "Cache timeout for maintenance type notifications in UsersMaintenanceNotificationViewset",
    },
    {
        "name": "MAINTENANCE_NOTIFICATIONS_CACHE_KEY",
        "default": "current_maintenance_notifications",
        "description": "Maintenance notifications list cache key.",
    },
    {
        "name": "TIME_BUFFER_FOR_CURRENT_MAINTENANCE_API_QUERY",
        "default": 900,
        "description": "When list maintenance notifications api is called with current=true in query param, "
        "maintenance notification within range of now - TIME_BUFFER_FOR_CURRENT_MAINTENANCE_API_QUERY, "
        "now  + TIME_BUFFER_FOR_CURRENT_MAINTENANCE_API_QUERY is returned. Maintenance notification "
        "cannot be created if it overlaps with annother maintenance notification in buffer.",
    },
)



# Generated by Django 4.2.4 on 2023-11-20 13:51

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0009_auto_20231108_0658"),
    ]

    operations = [
        migrations.AddField(
            model_name="djangoprojectbasenotification",
            name="extra_data",
            field=models.JSONField(null=True),
        ),
    ]



import uuid

from typing import List

import svgwrite
import swapper

from django.conf import settings
from django.contrib.auth import user_logged_in
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.core.validators import validate_comma_separated_integer_list
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _, pgettext_lazy
from rest_framework.exceptions import ErrorDetail, PermissionDenied
from taggit.models import GenericTaggedItemBase, TagBase

from django_project_base.base.fields import HexColorField


class BaseProject(models.Model):
    name = models.CharField(max_length=80, null=False, blank=False, db_index=True, verbose_name=_("Name"))
    slug = models.SlugField(max_length=80, null=False, blank=False, unique=True, verbose_name=_("Slug"))
    description = models.TextField(null=True, blank=True, verbose_name=_("Description"))
    logo = models.FileField(null=True, blank=True, verbose_name=_("Logo"))
    owner = parent = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Profile"), on_delete=models.CASCADE, verbose_name=_("Owner")
    )

    class Meta:
        abstract = True


class Project(BaseProject):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "Project")


class BaseProfile(User):
    """
    User profile. We start with some easy common settings
    """

    bio = models.TextField(max_length=500, null=True, blank=True, verbose_name=_("Bio"))
    phone_number = models.CharField(max_length=20, null=True, blank=True, verbose_name=_("Phone number"))
    # This one will list all supported languages
    language = models.CharField(max_length=10, null=True, blank=True, verbose_name=_("Language"))
    # This one will list all supported themes
    theme = models.CharField(max_length=10, null=True, blank=True, verbose_name=_("Theme"))
    avatar = models.FileField(null=True, blank=True, verbose_name=_("Avatar"))
    reverse_full_name_order = models.BooleanField(null=True, blank=True, verbose_name=_("Reverse full name order"))
    delete_at = models.DateTimeField(null=True, blank=True, verbose_name=_("Delete profile at"))
    password_invalid = models.BooleanField(default=False)

    def get_full_name(self, reverse_order: bool = False):
        # First we establish the natural order for this user
        if self.reverse_full_name_order is None:
            reversed_order = getattr(settings, "PROFILE_REVERSE_FULL_NAME_ORDER", False)
        else:
            reversed_order = self.reverse_full_name_order

        # Then we reverse this if requested for the method
        reversed_order = reversed_order ^ reverse_order

        first_name = self.first_name
        last_name = self.last_name
        if not (first_name or last_name):
            # if neither first nor last name were provided, we use the email address
            first_name = self.email

        if reversed_order:
            first_name, last_name = last_name, first_name

        return f"{first_name} {last_name}".strip()

    full_name = property(lambda self: self.get_full_name())
    full_name_reverse = property(lambda self: self.get_full_name(True))

    def get_users(self):
        objects_all = BaseProfile.objects
        return objects_all

    users = property(lambda self: self.get_users())

    def __str__(self):
        return self.full_name

    class Meta:
        abstract = True


class Profile(BaseProfile):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "Profile")


# noinspection PyUnusedLocal
@receiver(post_save, sender=User)
def create_user_profile(sender, instance, created, **kwargs):
    if created:
        profile = swapper.load_model("django_project_base", "Profile")
        profile(user_ptr=instance).save_base(raw=True)


class BaseProjectMember(models.Model):
    project = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Project"),
        on_delete=models.CASCADE,
        verbose_name=_("Project"),
        related_name="members",  # the name is just reversed: you seek members from the project
    )
    member = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Profile"),
        on_delete=models.CASCADE,
        verbose_name=_("Owner"),
        related_name="projects",  # the name is just reversed: you seek projects this member belongs to
    )

    @property
    def project_members_excluded_fields(self):
        return "id", "member", "project"

    @property
    def project_members_fields(self):
        fields_club = swapper.load_model("django_project_base", "ProjectMember")._meta.fields
        return [field for field in fields_club if field.name not in self.project_members_excluded_fields]

    @property
    def project_members_fields_names(self):
        return [field.name for field in self.project_members_fields]

    # role = models.ForeignKey()  # TODO: we don't have role support yet

    class Meta:
        abstract = True


class ProjectMember(BaseProjectMember):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "ProjectMember")


class BaseTag(TagBase):
    color = HexColorField(verbose_name=_("Color"))
    # icon if char for now # todo: what will be used for icon
    icon = models.CharField(max_length=10, null=True, blank=True, verbose_name=_("Icon"))
    project = models.ForeignKey(
        settings.DJANGO_PROJECT_BASE_PROJECT_MODEL,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        verbose_name=_("Project"),
    )
    name = models.CharField(verbose_name=pgettext_lazy("A tag name", "name"), unique=False, max_length=100)
    slug = models.SlugField(
        verbose_name=pgettext_lazy("A tag slug", "slug"), unique=False, max_length=100, allow_unicode=True
    )

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = self.slugify(self.name)
        self.full_clean()
        super().save(*args, **kwargs)

    @staticmethod
    def get_background_svg_for_tags(tags: List["BaseTag"]) -> str:
        dwg = svgwrite.Drawing()
        dwg.viewbox(width=200, height=200)
        tricolor_gradient = dwg.linearGradient((0, 0), (1, 1))
        dwg.defs.add(tricolor_gradient)
        color_added_flag: bool = True
        for i, t in enumerate(tags):
            if t.color:
                tricolor_gradient.add_stop_color(i / len(tags), t.color)
                t.color = True
        if not color_added_flag:
            return ""
        dwg.add(dwg.rect((10, 70), (50, 50), fill=tricolor_gradient.get_paint_server(default="currentColor")))
        return '<?xml version="1.0" encoding="utf-8" ?>\n' + dwg.tostring()

    class Meta(TagBase.Meta):
        abstract = True
        unique_together = [
            ["project", "name"],
            ["project", "slug"],
        ]


class DpbTaggedItemThrough(GenericTaggedItemBase):
    tag = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Tag"),
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_items",
    )
    # main_object actually states whether this tag is auto-generated. Used e.g. in UserGroups where a tag is always
    #   generated for a user group. The auto-generated ones have this flag set.
    main_object = models.BooleanField(default=False)

    class Meta(GenericTaggedItemBase.Meta):
        abstract = True


class Tag(BaseTag):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "Tag")


class BaseMergeUserGroup(models.Model):
    users = models.CharField(max_length=1024, null=False, validators=(validate_comma_separated_integer_list,))
    created_by = models.PositiveIntegerField()
    project = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Project"), on_delete=models.CASCADE, null=False
    )

    class Meta:
        abstract = True


# TODO: not sure this is even needed? why would anyone want to override this particular table?
class MergeUserGroup(BaseMergeUserGroup):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "MergeUserGroup")


class ProjectSettingsQs(models.query.QuerySet):
    def delete(self):
        raise PermissionDenied


class BaseProjectSettings(models.Model):
    VALUE_TYPE_INTEGER = "integer"
    VALUE_TYPE_FLOAT = "float"
    VALUE_TYPE_BOOL = "bool"
    VALUE_TYPE_CHAR = "char"
    # custom means that there's a custom editor for this setting, so it needs to be hidden from the general editor
    VALUE_TYPE_CUSTOM = "custom"

    VALUE_TYPE_CHOICES = (
        (VALUE_TYPE_INTEGER, _("Whole number")),
        (VALUE_TYPE_FLOAT, _("Decimal number")),
        (VALUE_TYPE_BOOL, _("True/False")),
        (VALUE_TYPE_CHAR, _("String")),
        (VALUE_TYPE_CUSTOM, _("Custom")),
    )

    objects = ProjectSettingsQs.as_manager()

    value_validators = {
        VALUE_TYPE_INTEGER: lambda val: models.IntegerField().to_python(val),
        VALUE_TYPE_FLOAT: lambda val: models.FloatField().to_python(val),
        VALUE_TYPE_BOOL: lambda val: models.BooleanField().to_python(val),
        VALUE_TYPE_CHAR: lambda val: models.TextField().to_python(val),
        VALUE_TYPE_CUSTOM: lambda val: val,
    }

    @property
    def python_value(self):
        return self.value_validators[self.value_type](self.value)

    @property
    def python_pending_value(self):
        return self.value_validators[self.value_type](self.pending_value)

    def clean(self):
        validator = self.value_validators[self.value_type]
        try:
            validator(self.value)
        except ValidationError as ve:
            from rest_framework.serializers import ValidationError as DrfValidationError

            exc = DrfValidationError()
            setattr(exc, "detail", ErrorDetail(next(iter(list(ve.params.keys()))), ve.messages))
            setattr(exc, "model-validation", True)
            raise exc
        super().clean()

    name = models.CharField(max_length=80, null=False, blank=False, db_index=True, verbose_name=_("Name"))
    description = models.CharField(max_length=120, null=False, blank=False, verbose_name=_("Description"))
    value = models.CharField(max_length=320, null=False, blank=False, verbose_name=_("Value"))
    value_type = models.CharField(choices=VALUE_TYPE_CHOICES, null=False, blank=False, max_length=10)

    reserved = models.BooleanField(null=False, default=False)

    project = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Project"), on_delete=models.CASCADE, null=False
    )

    pending_value = models.CharField(max_length=320, null=True, blank=True, verbose_name=_("Pending value"))
    action_required = models.BooleanField(default=False, null=True, blank=True)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.full_clean()
        validator = self.value_validators[self.value_type]
        self.value = validator(self.value)
        if self.pending_value is not None:
            self.pending_value = validator(self.pending_value)
        if self.action_required:
            from django_project_base.base.event import ProjectSettingActionRequiredEvent

            ProjectSettingActionRequiredEvent(user=None).trigger(payload=self)
        super().save(force_insert, force_update, using, update_fields)

    def delete(self, using=None, keep_parents=False):
        raise PermissionDenied

    class Meta:
        unique_together = [
            ["project", "name"],
        ]
        abstract = True


class ProjectSettings(BaseProjectSettings):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "ProjectSettings")


@receiver(user_logged_in)
def user_logged_in(*args, **kwargs):
    from django_project_base.account.service.merge_users_service import MergeUsersService

    MergeUsersService().handle(**kwargs)


class BaseInvite(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, verbose_name=_("Id"))
    email = models.CharField(max_length=255, verbose_name=_("eMail"))

    role = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Role"), null=True, on_delete=models.SET_NULL
    )

    text = models.TextField(verbose_name=_("Invitation message"), null=True, blank=False)

    invited_by = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Profile"),
        on_delete=models.CASCADE,
        related_name="project_user_invites",
    )
    accepted = models.DateTimeField(auto_now=False, null=True, blank=True)

    project = models.ForeignKey(
        swapper.get_model_name("django_project_base", "Project"), on_delete=models.CASCADE, null=False
    )

    class Meta:
        abstract = True
        unique_together = [
            ["project", "email"],
        ]


class Invite(BaseInvite):
    class Meta:
        swappable = swapper.swappable_setting("django_project_base", "Invite")



from enum import Enum
from typing import Optional, Union

from django.conf import Settings


class NotificationType(Enum):
    MAINTENANCE = "maintenance"
    STANDARD = "standard"


class NotificationLevel(Enum):
    SUCCESS = "success"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class ChannelIdentifier(Enum):
    MAIL = 0
    SMS = 1
    WEBSOCKET = 3

    @staticmethod
    def supported_channels() -> tuple:
        from django_project_base.notifications.base.channels.mail_channel import MailChannel
        from django_project_base.notifications.base.channels.sms_channel import SmsChannel

        return MailChannel(), SmsChannel()

    @staticmethod
    def channel(
        identifier: Union[int, str],
        settings: Optional[Settings] = None,
        project_slug: Optional[str] = None,
        ensure_dlr_user=True,
    ) -> Optional["Channel"]:  # noqa: F821
        channel = next(
            iter(filter(lambda c: c.name == identifier or c.id == identifier, ChannelIdentifier.supported_channels())),
            None,
        )
        if channel:
            channel.provider = channel._find_provider(settings=settings, setting_name=channel.provider_setting_name)
            channel.provider.ensure_credentials(settings=settings)
            channel.provider.ensure_dlr_user(project_slug) if ensure_dlr_user else False

            return channel
        return None



from django.urls import include, path, re_path
from django.views.i18n import JavaScriptCatalog
from drf_spectacular.settings import SPECTACULAR_DEFAULTS

from django_project_base.router import django_project_base_router
from django_project_base.views import browser_update_script

SPECTACULAR_DEFAULTS["TITLE"] = "Rest documentation"
SPECTACULAR_DEFAULTS["DESCRIPTION"] = "Api documentation"
SPECTACULAR_DEFAULTS["VERSION"] = "0.0.1"
SPECTACULAR_DEFAULTS["PREPROCESSING_HOOKS"] = ["django_project_base.router.filter_rest_documentation_endpoints"]

SPECTACULAR_DEFAULTS["POSTPROCESSING_HOOKS"] = ["django_project_base.router.postprocess_rest_documentation"]

urlpatterns = [
    path("jsi18n/", JavaScriptCatalog.as_view(), name="javascript-catalog"),
    re_path("browser-update/js-script", browser_update_script, name="browser-update-script"),
    re_path(r"", include(django_project_base_router.urls)),
]



import socket

import swapper
from django.urls import reverse
from rest_framework import status
from rest_framework.response import Response

from django_project_base.notifications.email_notification import EMailNotification
from django_project_base.notifications.models import DjangoProjectBaseMessage, DjangoProjectBaseNotification
from django_project_base.notifications.tests.notifications_transaction_test_case import NotificationsTransactionTestCase


class TestListMails(NotificationsTransactionTestCase):
    number_of_emails = 10

    def setUp(self) -> None:
        super().setUp()
        self.mail_data = dict(
            title="test list mail",
            message="test list mail content",
            recipients=[self.test_user.pk],
            sender="support@velis.si",
        )
        for i in range(0, self.number_of_emails):
            EMailNotification(
                message=DjangoProjectBaseMessage(
                    subject=self.mail_data["title"], body=self.mail_data["message"], footer=""
                ),
                raw_recipents=[
                    self.test_user.pk,
                ],
                project=swapper.load_model("django_project_base", "Project").objects.first().slug,
                recipients=[
                    self.test_user.pk,
                ],
                content_entity_context="example_email" "",
            ).send()

    def test_list_internal_mails(self):
        response: Response = self.api_client.get(
            reverse("notification-list"),
            format="json",
            HTTP_HOST=socket.gethostname().lower(),
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self._login_to_api_client_with_test_user()
        response: Response = self.api_client.get(
            reverse("notification-list"),
            format="json",
            HTTP_HOST=socket.gethostname().lower(),
            HTTP_CURRENT_PROJECT=swapper.load_model("django_project_base", "Project").objects.first().slug,
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(1, len(response.data))

        self.assertEqual(self.number_of_emails,
                         DjangoProjectBaseNotification.objects.get(pk=response.data[0]["id"]).counter)



import datetime

from typing import Optional

from django.conf import settings
from django.db import transaction
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view, OpenApiParameter
from dynamicforms.serializers import ModelSerializer as DFSerializer
from dynamicforms.viewsets import ModelViewSet
from rest_framework import fields, status
from rest_framework.authentication import BasicAuthentication, SessionAuthentication, TokenAuthentication
from rest_framework.decorators import action
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer, Serializer as RestFrameworkSerializer

from django_project_base.notifications.base.enums import NotificationType
from django_project_base.notifications.maintenance_notification import MaintenanceNotification
from django_project_base.notifications.models import DjangoProjectBaseMessage, DjangoProjectBaseNotification
from django_project_base.notifications.utils import utc_now
from django_project_base.utils import get_pk_name


class NotificationAcknowledgedRequestSerializer(RestFrameworkSerializer):
    def __new__(cls, *args, **kwargs):
        new: "NotificationAcknowledgedRequestSerializer" = super().__new__(cls, *args, **kwargs)
        new.fields[get_pk_name(DjangoProjectBaseMessage)] = fields.UUIDField(
            required=True, allow_null=False, help_text="Notification identifier"
        )
        new.fields["acknowledged_identifier"] = fields.IntegerField(
            required=True,
            allow_null=False,
            help_text=_("Time interval identifying at what time notification was acknowledged by user"),
        )
        return new

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass


class MessageSerializer(DFSerializer):
    class Meta:
        model = DjangoProjectBaseMessage
        exclude = (get_pk_name(DjangoProjectBaseMessage),)


class MaintenanceNotificationSerializer(ModelSerializer):
    notification_acknowledged_data = fields.SerializerMethodField()
    message = MessageSerializer()

    def get_notification_acknowledged_data(self, notification: DjangoProjectBaseNotification) -> list:
        request: Optional[Request] = self.context.get("request")
        return request.session.get(settings.MAINTENANCE_NOTIFICATIONS_CACHE_KEY, {}).get(str(notification.pk), [])

    def create(self, validated_data) -> DjangoProjectBaseNotification:
        try:
            message: DjangoProjectBaseMessage = DjangoProjectBaseMessage.objects.create(**validated_data["message"])
            return MaintenanceNotification(delay=validated_data["delayed_to"], message=message).send()
        except AssertionError as ae:
            raise ValidationError(str(ae))
        except Exception as e:
            raise APIException(e)

    # todo: update and partial update, destroy

    def validate(self, attrs: dict):
        _type: Optional[str] = attrs.get("type")
        if _type and _type != NotificationType.MAINTENANCE.value:
            raise ValidationError({"type": "Only type %s allowed." % NotificationType.MAINTENANCE.value})
        time_delta: int = settings.TIME_BUFFER_FOR_CURRENT_MAINTENANCE_API_QUERY
        existing_maintenances: list = DjangoProjectBaseNotification.objects.filter(
            delayed_to__range=[attrs["delayed_to"] - time_delta, attrs["delayed_to"] + time_delta]
        )
        if bool(len(existing_maintenances)):
            raise ValidationError({"delayed_to": "Another maintenance is planned at this time"})
        return super().validate(attrs)

    class Meta:
        model = DjangoProjectBaseNotification
        exclude = (
            "required_channels",
            "sent_channels",
            "failed_channels",
            "recipients",
            "recipients_original_payload",
            "level",
            "sent_at",
            "type",
            "content_entity_context",
        )


@extend_schema_view(
    destroy=extend_schema(exclude=True),
    update=extend_schema(exclude=True),
    partial_update=extend_schema(exclude=True),
)
class UsersMaintenanceNotificationViewset(ModelViewSet):
    authentication_classes = [SessionAuthentication, BasicAuthentication, TokenAuthentication]
    permission_classes = [IsAuthenticated]

    def get_serializer_class(self):
        return MaintenanceNotificationSerializer

    def get_queryset(self):
        now: int = utc_now().timestamp()
        return DjangoProjectBaseNotification.objects.filter(
            type=NotificationType.MAINTENANCE.value,
            delayed_to__gt=now,
            delayed_to__lt=now + datetime.timedelta(hours=8).total_seconds(),
        )

    @extend_schema(request=MaintenanceNotificationSerializer(), description="Create maintenance notification.")
    @transaction.atomic
    def create(self, request: Request, *args, **kwargs) -> Response:
        return super().create(request, *args, **kwargs)

    @extend_schema(
        request=MaintenanceNotificationSerializer(many=True),
        description="List's maintenance notifications that are planned in future.",
        parameters=[
            OpenApiParameter(
                name="current",
                description="If present in request and if true, then this api "
                "returns only maintenance notification that is planned in range "
                "of now +/- TIME_BUFFER_FOR_CURRENT_MAINTENANCE_API_QUERY settings value.",
                required=False,
                type=bool,
            ),
        ],
    )
    def list(self, request: Request, *args, **kwargs) -> Response:
        if request.query_params.get("current", "False") in fields.BooleanField.TRUE_VALUES:
            now: float = utc_now().timestamp()
            time_delta: int = settings.TIME_BUFFER_FOR_CURRENT_MAINTENANCE_API_QUERY
            current_maintenance: Optional[DjangoProjectBaseNotification] = next(
                iter(
                    DjangoProjectBaseNotification.objects.filter(delayed_to__range=[now - time_delta, now + time_delta])
                ),
                None,
            )
            return Response(self.get_serializer(current_maintenance, many=False).data)
        read_notifications: dict = request.session.get(settings.MAINTENANCE_NOTIFICATIONS_CACHE_KEY, {})
        pk_name: str = get_pk_name(DjangoProjectBaseMessage)
        return Response(
            self.get_serializer(
                filter(lambda n: len(read_notifications.get(str(getattr(n, pk_name)), [])) < 3, self.get_queryset()),
                many=True,
            ).data
        )

    @extend_schema(description="Get single maintenance notification's data")
    def retrieve(self, request: Request, *args, **kwargs) -> Response:
        return super().retrieve(request, *args, **kwargs)

    @transaction.atomic
    def partial_update(self, request: Request, *args, **kwargs) -> Response:
        raise APIException(code=status.HTTP_501_NOT_IMPLEMENTED)
        return super().partial_update(request, *args, **kwargs)

    @transaction.atomic
    def update(self, request: Request, *args, **kwargs) -> Response:
        raise APIException(code=status.HTTP_501_NOT_IMPLEMENTED)
        return super().update(request, *args, **kwargs)

    @transaction.atomic
    def destroy(self, request: Request, *args, **kwargs) -> Response:
        raise APIException(code=status.HTTP_501_NOT_IMPLEMENTED)
        return super().destroy(request, *args, **kwargs)

    @extend_schema(
        request=NotificationAcknowledgedRequestSerializer(),
        description="Mark message as acknowledged by user making request.",
        responses={status.HTTP_201_CREATED: None},
    )
    @action(methods=["POST"], detail=False, url_path="acknowledged", url_name="acknowledged")
    def acknowledged(self, request: Request, **kwargs) -> Response:
        # todo: storage for acknowledged notifications should be configurable
        ser: NotificationAcknowledgedRequestSerializer = NotificationAcknowledgedRequestSerializer(data=request.data)
        ser.is_valid(raise_exception=True)
        if settings.MAINTENANCE_NOTIFICATIONS_CACHE_KEY not in request.session:
            request.session[settings.MAINTENANCE_NOTIFICATIONS_CACHE_KEY] = {}
        read_messages: dict = request.session[settings.MAINTENANCE_NOTIFICATIONS_CACHE_KEY]
        notice_pk: str = ser.data[get_pk_name(DjangoProjectBaseMessage)]
        notice_diff: int = ser.data["acknowledged_identifier"]
        if notice_pk not in read_messages:
            read_messages[notice_pk] = []
        diffs: list = read_messages[notice_pk]
        diffs.append(notice_diff)
        diffs = list(set(diffs))
        read_messages[notice_pk] = diffs
        request.session[settings.MAINTENANCE_NOTIFICATIONS_CACHE_KEY] = read_messages
        return Response(status=status.HTTP_201_CREATED)



# Generated by Django 4.2.4 on 2023-08-18 14:42

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django_project_base.notifications.models
import uuid


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.DJANGO_PROJECT_BASE_PROJECT_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="SearchItems",
            fields=[
                ("content_type_id", models.PositiveIntegerField()),
                ("object_id", models.PositiveBigIntegerField()),
                ("name", models.CharField(max_length=256)),
                (
                    "id",
                    models.PositiveBigIntegerField(primary_key=True, serialize=False),
                ),
            ],
            options={
                "db_table": "search-items-table",
                "abstract": False,
                "managed": False,
            },
        ),
        migrations.CreateModel(
            name="DjangoProjectBaseMessage",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Id",
                    ),
                ),
                (
                    "subject",
                    models.TextField(blank=True, null=True, verbose_name="Subject"),
                ),
                ("body", models.TextField(verbose_name="Body")),
                (
                    "footer",
                    models.TextField(blank=True, null=True, verbose_name="Footer"),
                ),
                (
                    "content_type",
                    models.CharField(
                        choices=[("text/plain", "Plain text"), ("text/html", "Html")],
                        default="text/plain",
                        max_length=64,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="DjangoProjectBaseNotification",
            fields=[
                (
                    "locale",
                    models.CharField(blank=True, max_length=8, null=True, verbose_name="Locale"),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Id",
                    ),
                ),
                (
                    "level",
                    models.CharField(
                        choices=[
                            ("success", "Success"),
                            ("info", "Info"),
                            ("warning", "Warning"),
                            ("error", "Error"),
                        ],
                        max_length=16,
                        verbose_name="Level",
                    ),
                ),
                (
                    "required_channels",
                    models.CharField(
                        blank=True,
                        max_length=32,
                        null=True,
                        verbose_name="Required channels",
                    ),
                ),
                (
                    "sent_channels",
                    models.CharField(
                        blank=True,
                        max_length=32,
                        null=True,
                        verbose_name="Sent channels",
                    ),
                ),
                (
                    "failed_channels",
                    models.CharField(
                        blank=True,
                        max_length=32,
                        null=True,
                        verbose_name="Failed channels",
                    ),
                ),
                (
                    "created_at",
                    models.BigIntegerField(
                        default=django_project_base.notifications.models.integer_ts,
                        editable=False,
                        verbose_name="Created at",
                    ),
                ),
                (
                    "sent_at",
                    models.BigIntegerField(blank=True, null=True, verbose_name="Sent at"),
                ),
                (
                    "delayed_to",
                    models.BigIntegerField(blank=True, null=True, verbose_name="Delayed to"),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("maintenance", "Maintenance"),
                            ("standard", "Standard"),
                        ],
                        default="standard",
                        max_length=16,
                        verbose_name="Type",
                    ),
                ),
                (
                    "recipients",
                    models.CharField(
                        max_length=2048,
                        null=True,
                        validators=[django.core.validators.int_list_validator],
                    ),
                ),
                ("exceptions", models.TextField(null=True)),
                ("content_entity_context", models.TextField()),
                ("counter", models.SmallIntegerField(default=1)),
                ("recipients_original_payload", models.CharField(max_length=2048)),
                (
                    "recipients_original_payload_search",
                    models.TextField(db_index=True, null=True),
                ),
                (
                    "message",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="notifications.djangoprojectbasemessage",
                        verbose_name="Message",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.apps import AppConfig

from django_project_base.settings import set_django_project_base_settings


class DjangoProjectBaseConfig(AppConfig):
    name = "django_project_base"

    verbose_name = "Django Project Base"

    def ready(self):
        set_django_project_base_settings()
        import django_project_base.base.signals  # noqa: F401



import functools

from django_project_base.profiling.middleware import ProfileRequest


def function_profiler(name=None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            path_info = name or f"{func.__module__}.{func.__qualname__}"
            params = ""
            if args:
                params += f"args={args} "
            for key, value in kwargs.items():
                params += f"{key}={value} "
            with ProfileRequest(
                {"REQUEST_METHOD": "FUNCTION", "HTTP_HOST": "", "QUERY_STRING": params, "PATH_INFO": path_info},
                func,
                args,
                kwargs,
            ) as pr:
                return pr.response

        return wrapper

    return decorator



from typing import Any, cast, Dict, Optional, Union

import swapper

from django.core.validators import RegexValidator
from django.db.models import fields, Q, QuerySet
from django.utils.translation import gettext_lazy as _
from dynamicforms import fields as df_fields

from django_project_base.base.filter_to_model import filter_queryset_to_project


class HexColorField(fields.CharField):
    def __init__(self, *args, **kwargs):
        if "max_length" not in kwargs:
            kwargs["max_length"] = 7
        super().__init__(*args, **kwargs)
        self.validators.append(
            RegexValidator(
                regex="^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$",
                message=_("Value is not a hex color (e.g #88f or #8080ff)"),
            )
        )


class UserRelatedField(df_fields.PrimaryKeyRelatedField):
    def __init__(
        self,
        queryset_filter: Optional[Union[Q, Dict[str, Any]]] = None,
        queryset_exclude: Optional[Union[Q, Dict[str, Any]]] = None,
        **kwargs,
    ):
        kwargs["url_reverse"] = "profile-base-project-list"
        kwargs["query_field"] = "full_name"
        kwargs["additional_parameters"] = dict(select=1)
        kwargs["text_field"] = "full_name"
        kwargs["value_field"] = "id"
        super().__init__(**kwargs)
        self.queryset_filter = queryset_filter or {}
        self.queryset_exclude = queryset_exclude or {}
        self.filter_selected_project = True  # ImpersonateUserIdField overrides this (or any other; as needed)

    def get_queryset(self):
        # TODO This needs to be amended together with members editor such that it will be possible to specify
        #  in settings.py how to filter and sort project members
        qs = cast(QuerySet, swapper.load_model("django_project_base", "Profile").objects)

        if self.filter_selected_project:
            try:
                request = self.context.get("request", None)
                qs = filter_queryset_to_project(qs, "projects__project", request.selected_project)
            except AttributeError:
                # if the middleware setting current project is not even set (request.selected_project will except),
                #  we just show all users
                pass

        qs = qs.exclude(**self.queryset_exclude if isinstance(self.queryset_exclude, dict) else self.queryset_exclude)
        qs = qs.filter(**self.queryset_filter if isinstance(self.queryset_filter, dict) else self.queryset_filter)

        return qs.all()

    def display_value(self, instance):
        if not instance:
            return ""

        return str(instance)



import time

from django.core.cache import cache

from django_project_base.caching import CacheCounter


class ObjectLockTimeout(Exception):
    """ Exception raised when a lock timeout is exceeded."""


class NoTimeoutCheck(Exception):
    """
    Exception raised when timeout is set, but not checked.
    Example:
    with CacheLock as cl:
        cl()  # this calls timeout check
        ...
        ...
    """


class CacheLock(object):
    is_waiting = False
    waiting_counter = None

    def __init__(self, name, timeout=0, silence_object_lock_timeout=False, stats_name=None):
        """
        :param name: name of the cache
        :param timeout: timeout in seconds.
            If timeout is 0, no timeout.
            If timeout is -1, code in with block will not be executed if another process is already executing it
        :param silence_object_lock_timeout: True means no exception will be raised if object lock timeout is exceeded.

        If timeout is != 0, timeout check must be called first thing in with block.
        with CacheLock(name='mycache', timeout=1) as cl:
            cl()
            ...
            ...
            rest of code in with block
            ...
            ...
        """
        if timeout is None or (timeout < 0 and timeout != -1):
            raise Exception("Invalid timeout value")

        self.name = "CacheLock." + name
        self.stats_name = "CacheLock." + (stats_name or name)
        self.timeout = timeout
        self.silence_object_lock_timeout = silence_object_lock_timeout
        self.raise_timeout_exception = False
        self.timeout_checked = False

    # noinspection PyMethodMayBeStatic
    def append_waiting_key(self, key):
        if not cache.has_key(f"Inserted.{key}"):
            cache.set(f"Inserted.{key}", True, timeout=None)
            from django_project_base.caching.cache_queue import CacheQueue

            CacheQueue.get_cache_queue("CacheLockKeys", timeout=None).rpush(key)

    def set_waiting(self, is_waiting):
        if is_waiting:
            if not self.is_waiting:
                self.is_waiting = True
                key = f"Waiting.{self.stats_name}"
                self.waiting_counter = CacheCounter(key, timeout=None)
                self.waiting_counter.incr()
                self.append_waiting_key(key)
        elif self.is_waiting and not is_waiting:
            self.waiting_counter.incr(step=-1)

    def __enter__(self):
        try:
            start_time = time.time()
            while True:
                res = CacheCounter(self.name, timeout=None).incr()
                if res == 1:
                    self.set_waiting(False)
                    break
                elif self.timeout == -1:
                    self.raise_timeout_exception = True
                    break
                self.set_waiting(True)
                time.sleep(0.1)
                if 0 < self.timeout < time.time() - start_time:
                    self.raise_timeout_exception = True
                    self.set_waiting(False)
                    break
            return self
        except Exception as e:  # pragma: no cover
            self.set_waiting(False)
            raise e

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.timeout != 0 and not self.timeout_checked:
            cache.delete(self.name)
            raise NoTimeoutCheck()

        if exc_type is ObjectLockTimeout:
            return self.silence_object_lock_timeout
        cache.delete(self.name)

    def __call__(self, *args, **kwargs):
        self.timeout_checked = True
        if self.raise_timeout_exception:
            raise ObjectLockTimeout()



from django import template
from hijack.templatetags.hijack_tags import is_hijacked

register = template.Library()


@register.filter
def is_request_hijacked(request):
    return is_hijacked(request)



"""
query_tracker

this submodule allows for tracking SQL queries and outputting them to logs

activate by setting database ENGINE to "django_project_base.query_tracker"

it will also check for these additional parameters:

"TRACKED_ENGINE": mandatory. what was "ENGINE" before. tracker will instantiate this engine with the rest of options

"TRACKER_LOGGER_NAME": default None. name of logger to use for logging SQL requests.
  See https://docs.djangoproject.com/en/dev/topics/logging/

"TRACKER_LOGGER_LEVEL": default logging.WARNING. if a falsy value is given, it is reset to DEBUG. Needs to be int.
  Also see django logging docs or python logging module docs
  Note: when manage.py test has been run, the default becomes logging.DEBUG. Default logger will ignore the logging

"TRACKER_FILTER_STACK": default ("site-packages", "query_tracker", "/python3", "JetBrains").
  Whether we should filter the stack to only show "relevant" stack code points, i.e. "our own" code.
  set to empty tuple to NOT filter the stack or specify a tuple of strings that should not be in code path


Example DATABASES configuration from settings.py:
DATABASES = {
    "default": {
        "ENGINE": "django_project_base.query_tracker",
        "TRACKED_ENGINE": "django.db.backends.sqlite3",
        "TRACKER_LOGGER_LEVEL": logging.INFO,
        "TRACKER_FILTER_STACK", ("site-packages", "query_tracker", "/python3", "JetBrains"),
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

Output for each query consists of:
* request path (if it could be found, requires "django_project_base.base.UrlVarsMiddleware" to be installed
* query itself, prefixed by execution time in ms
* stack trace that led to the query
"""

import importlib
import logging
import sys
import time
import traceback

from typing import Optional, Tuple

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.utils import CursorWrapper

from django_project_base.base.middleware import get_current_request


def filter_stack(filter_stack: Tuple[str] = tuple()):
    # Get the current stack trace
    stack = traceback.extract_stack()

    # Filter out the entries containing ".venv"
    if filter_stack:
        filtered_stack = filter(
            lambda entry: not any(s in entry.filename for s in filter_stack),
            stack,
        )
    else:
        filtered_stack = stack

    # Format the filtered stack trace
    formatted_stack = traceback.format_list(filtered_stack)

    return "".join(formatted_stack)


def quote_strings(val):
    if isinstance(val, str):
        return f"'{val}'"
    return val


class StackTraceCursorWrapper(CursorWrapper):
    def __init__(self, logger_name: Optional[str], logger_level: int, filter_stack: Tuple[str], *args, **kwds):
        self.logger = logging.getLogger(logger_name)
        self.logger_level = logger_level
        self.filter_stack = filter_stack
        super().__init__(*args, **kwds)

    def execute(self, sql, params=None):
        log_lines = []
        try:
            request = get_current_request()
            log_lines.append(" ".join(["request path", request.path]))
        except:
            pass
        log_lines.append(" ".join(["sql", sql % tuple(map(quote_strings, params)) if params else sql]))
        log_lines.append(filter_stack(self.filter_stack))
        tim = time.time()
        res = super().execute(sql, params)
        tim = (time.time() - tim) * 1000
        log_lines.append(" ".join(["sql", f"{tim:.2f}ms", sql % tuple(map(quote_strings, params)) if params else sql]))
        self.logger.log(self.logger_level, "\n".join(log_lines))
        return res

    def executemany(self, sql, param_list):
        log_lines = []
        try:
            request = get_current_request()
            log_lines.append(" ".join(["request path", request.path]))
        except:
            pass
        log_lines.append(" ".join(["sql", sql]))
        log_lines.append(filter_stack())
        self.logger.log(self.logger_level, "\n".join(log_lines))
        return super().executemany(sql, param_list)


class DatabaseWrapper(BaseDatabaseWrapper):
    def __new__(cls, settings_dict, *args, **kwargs):
        assert "TRACKED_ENGINE" in settings_dict
        testing = len(sys.argv) > 1 and sys.argv[1] == "test" and "manage.py" in sys.argv[0]
        default_level = logging.DEBUG if testing else logging.WARNING

        tracked_engine = settings_dict["TRACKED_ENGINE"] + ".base"
        logger_name = settings_dict.get("TRACKER_LOGGER_NAME", None)
        logger_level = settings_dict.get("TRACKER_LOGGER_LEVEL", default_level) or logging.DEBUG
        filter_stack = settings_dict.get(
            "TRACKER_FILTER_STACK", ("site-packages", "query_tracker", "/python3", "JetBrains")
        )
        assert isinstance(logger_level, int)

        module = importlib.import_module(tracked_engine)
        DBW = getattr(module, "DatabaseWrapper")

        class CDBW(DBW):
            def create_cursor(self, name=None):
                cursor = super().create_cursor(name)
                return StackTraceCursorWrapper(logger_name, logger_level, filter_stack, cursor, self)

        return CDBW(settings_dict, *args, **kwargs)
